/* makefont.c - Converts a text font for Dos, Linux, or Windows */

// Author: Byron Stanoszek
//
// Version History:
//  1.0  Aug 1, 2000, Initial version.
//       Converts a bitmap cell font into a Windows .fnt file.
//       
//  1.1  Feb 25, 2001.
//       Converts a plaintext font format to Dos, Linux, or Windows.

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#define COPYRIGHT "Copyright (C) 2001 by Byron Stanoszek"
#define VERSION   "1.1, 02/25/2001"

int height;
char font[256][32];
char list[256];

enum font_types { LINUX, DOS, WINDOWS };


/* Writes an integer in little-endian format */
void putnum(FILE *f, int sz, int num)
{
  fputc(num, f);
  if(sz > 1)
    fputc(num >> 8, f);
  if(sz > 2)
    fputc(num >> 16, f);
  if(sz > 3)
    fputc(num >> 24, f);
  while(sz > 4)
    fputc(0, f), sz--;
}

/* Read a font from a .txt file */
void read_font(char *filename)
{
  FILE *f;
  char buf[64];
  int i, j, chr=-1;

  if(!(f=fopen(filename, "r"))) {
    perror(filename);
    exit(1);
  }

  fgets(buf, 64, f);
  height=atoi(buf);
  if(height < 4 || height > 32) {
    fprintf(stderr, "Error: Bad height in %s: %d\n", filename, height);
    exit(1);
  }

  while(fgets(buf, 60, f)) {
    if(*buf < '0' || *buf > '9') {
      fprintf(stderr, "Error: Expecting character number %d in %s.\n",
              chr+1, filename);
      exit(1);
    }

    chr=atoi(buf);
    if(chr < 0 || chr > 255) {
      fprintf(stderr, "Error: Bad character number %d.\n", chr);
      exit(1);
    }
    if(list[chr]) {
      fprintf(stderr, "Multiple character %d defined.\n", chr);
      for(i=0;i<height;i++)
        font[chr][i]=0;
    }
    list[chr]=1;

    for(i=0;i<height;i++) {
      fgets(buf, 60, f);
      for(j=0;j<8;j++)
        if(buf[j*2+1] != ' ')
          font[chr][i] |= (1 << (7-j));
    }
  }
  fclose(f);

  for(chr=0;chr<256;chr++)
    if(!list[chr])
      fprintf(stderr, "Character %d not defined.\n", chr);
}


/** Linux Section **/

void write_linux(FILE *g)
{
  int i, j;

  /* linux font header */
  fputc(54, g);
  fputc(4, g);
  fputc(0, g);
  fputc(height, g);

  for(i=0;i<256;i++)
    for(j=0;j<height;j++)
      fputc(font[i][j], g);
}


/** Dos Section **/

char dos_header[51]={
  "\xb8\x00\x05\xcd\x10\xb4\x0f\xcd"
  "\x10\xbf\x2d\x01\x8a\x25\x3a\xc4"
  "\x74\x06\x8a\xc4\x32\xe4\xcd\x10"
  "\xbd\x36\x01\xbf\x33\x01\x8a\x3d"
  "\xb9\x00\x01\x33\xd2\x32\xdb\xb8"
  "\x10\x11\xcd\x10\xc3\x03\x02\x01"
  "\x01\x04\x41"};

void write_dos(FILE *g)
{
  int i, j;

  /* write 16-bit dos assembly header */
  fwrite(&dos_header, 51, 1, g);

  /* font height */
  fputc(height, g);

  /* number of characters */
  putnum(g, 2, 256);

  for(i=0;i<256;i++)
    for(j=0;j<height;j++)
      fputc(font[i][j], g);
}


/** Windows Section **/

char copyright[60]="Created using the TinyMARE Font Converter! (www.winds.org)";
char fontname[256]="TinyMARE";
char devicename[256]="";

void write_windows(FILE *g, int width)
{
  size_t offset[5], table;
  int i, j, num, cellbytes;

  /* Write main font structure */
  putnum(g, 2, 0x0300);		// Version

  offset[0]=ftell(g);
  putnum(g, 4, 0);		// Total file size (to be filled in later)
  fwrite(copyright, 60, 1, g);	// Copyright string
  putnum(g, 2, 0);		// Type of font file. 0=Raster, 1=Vector.
  putnum(g, 2, width*10);	// Point size at which char looks best.
  putnum(g, 2, height);		// Vertical resolution, dpi.
  putnum(g, 2, width*2);	// Horizontal resolution, dpi.
  putnum(g, 2, height-2);	// Vertical ascent to baseline of font.
  putnum(g, 2, 0);		// Internal vertical leading for accent marks.
  putnum(g, 2, 0);		// External vertical leading between rows.
  putnum(g, 1, 0);		// 1=Italic
  putnum(g, 1, 0);		// 1=Underline
  putnum(g, 1, 0);		// 1=Strike-out
  putnum(g, 2, 400);		// Character weight (1-1000)
  putnum(g, 1, 0);		// Character set
  putnum(g, 2, width);		// Font width
  putnum(g, 2, height);		// Font height
  putnum(g, 1, 0x30);		// Font pitch/family, 0x30=Fixed width
  putnum(g, 2, width);		// Average width of character (or size of 'X')
  putnum(g, 2, width);		// Maximum width of any character in font
  putnum(g, 1, 0);		// First character# in font
  putnum(g, 1, 255);		// Last character# in font
  putnum(g, 1, ' ');		// Default character when out of range.
  putnum(g, 1, ' ');		// Default space/padding character.
  putnum(g, 2, 512);		// Number of bytes/row in bitmap.

  offset[1]=ftell(g);
  putnum(g, 4, 0);		// Offset in file to device name.
  putnum(g, 4, 0);		// Offset in file to facename of font.
  putnum(g, 4, 0);		// Offset to machine address of bitmap. Unset.
  putnum(g, 4, 0);		// Offset in file to beginning of bitmaps.
  putnum(g, 1, 0);		// Reserved.

  putnum(g, 4, 1);		// Glyph type. 1=Fixed pitch.
  putnum(g, 2, 0);		// A space. Unset.
  putnum(g, 2, 0);		// B space. Unset.
  putnum(g, 2, 0);		// C space. Unset.
  putnum(g, 4, 0);		// Offset to color index. DIB format w/out hdr.
  putnum(g, 16, 0);		// Reserved.

  /* Generate character table for #chars + 1 */
  num=257;
  table=ftell(g)+(num*6);
  cellbytes=((width+7)/8)*height;

  for(i=0;i<num;i++) {
    putnum(g, 2, width);
    putnum(g, 4, table);
    table+=cellbytes;
  }

  /* Write pixel data to font file */
  num--;
  offset[2]=ftell(g);
  for(i=0;i<num;i++) {
    for(j=0;j<height;j++)
      fputc(font[i][j], g);
    if(width == 9)
      for(j=0;j<height;j++)
        putnum(g, 1, (i >= 192 && i < 224 && (font[i][j] & 0x01))?0x80:0x00);
  }

  /* Write "blank" character */
  for(i=0;i<cellbytes;i++)
    fputc(0, g);

  /* Write name of font, null terminated */
  offset[3]=ftell(g);
  fputs(fontname, g);
  fputc(0, g);

  /* Device name (unused) */
  offset[4]=ftell(g);
  fputs(devicename, g);
  fputc(0, g);

  /* End of file. */
  offset[5]=ftell(g);

  /* Fulfill offsets */
  fseek(g, offset[0], SEEK_SET);
  putnum(g, 4, offset[5]);
  fseek(g, offset[1], SEEK_SET);
  putnum(g, 4, offset[4]);
  putnum(g, 4, offset[3]);
  putnum(g, 4, 0);
  putnum(g, 4, offset[2]);
}


/* Main Entry */
int main(int argc, char *argv[])
{
  FILE *g;
  int c=2, width=8, type=LINUX;

  if(argc < 3) {
    fprintf(stderr, "Linux/DOS/Windows Font Converter, version "VERSION"\n");
    fprintf(stderr, COPYRIGHT"\n\n");
    fprintf(stderr, "Usage: %s textfile -linux outfile\n", *argv);
    fprintf(stderr, "       %s textfile -dos outfile\n", *argv);
    fprintf(stderr, "       %s textfile -windows outfile fontname [width]\n",
            *argv);
    exit(1);
  } 

  if(argc > c && *argv[c] == '-') {
    if(!strcasecmp(argv[c]+1, "linux")) type=LINUX;
    else if(!strcasecmp(argv[c]+1, "dos")) type=DOS;
    else if(!strcasecmp(argv[c]+1, "windows")) type=WINDOWS;
    else {
      fprintf(stderr, "%s: unknown font format '%s'\n", argv[c]+1, *argv);
      exit(1);
    }
    c++;
  }

  read_font(argv[1]);

  if(argc > c) {
    if(!(g=fopen(argv[c], "wb"))) {
      perror(argv[c]);
      exit(1);
    }
    c++;
  } else {
    fprintf(stderr, "Error: No output file specified.\n");
    exit(1);
  }

  switch(type) {
  case LINUX:
    write_linux(g);
    break;
  case DOS:
    write_dos(g);
    break;
  case WINDOWS:
    if(argc > c)  /* Set font name */
      strncpy(fontname, argv[c++], 255);
    if(argc > c) {  /* Set width */
      width=atoi(argv[c++]);
      if(width != 8 && width != 9) {
        fprintf(stderr, "%s: font width can only be 8 or 9\n", *argv);
        exit(1);
      }
    }

    write_windows(g, width);
    break;
  }

  exit(0);
}
