/* comm/create.c */
/* commands that create new objects */

#include "externs.h"

int clone_depth;  /* Protection from @aclone recursion */


/* links an object's home or an exit's destination */
void do_link(dbref player, char *arg1, char *arg2)
{
  dbref thing, room=NOTHING;

  /* check restrictions */
  if((thing = match_thing(player, arg1, MATCH_EVERYTHING)) == NOTHING)
    return;
  if(Typeof(thing) != TYPE_ROOM || *arg2)
    if((room=match_thing(player,arg2,MATCH_EVERYTHING|MATCH_HOME)) == NOTHING)
      return;
  if(!can_link(player, thing, room))
    return;

  /* set the link */
  if(Typeof(thing) == TYPE_ROOM && room == thing)
    room=NOTHING;
  db[thing].link = room;

  /* display success message */
  if(Typeof(thing) == TYPE_ZONE)
    notify(player, tprintf("Origin Room set to %s.", unparse_object(player,
           room)));
  else if(Typeof(thing) == TYPE_ROOM) {
    if(room == NOTHING)
      notify(player, "Dropto removed.");
    else
      notify(player, tprintf("Dropto set to %s.",unparse_object(player,room)));
  } else
    notify(player, tprintf("%s linked to %s.", db[thing].name,
           unparse_object(player, room)));
}

int can_link(dbref player, dbref thing, dbref room)
{
  dbref *ptr;

  /* check powers */
  if(!controls(player, thing, POW_MODIFY)) {
    notify(player, "Permission denied.");
    return 0;
  } if(Guest(player)) {
    notify(player, "Guests cannot @link anything.");
    return 0;
  } if(RESTRICT_TELEPORT && !Builder(player)) {
    if(Typeof(thing) == TYPE_PLAYER) {
      notify(player, "Only builders may link players to other rooms.");
      return 0;
    } if(Typeof(thing) == TYPE_ROOM) {
      notify(player, "Only builders may set droptos on rooms.");
      return 0;
    } if(room == HOME) {
      notify(player, "Only builders may link exits to home.");
      return 0;
    } if(Typeof(thing) == TYPE_THING && mainroom(thing) != mainroom(room)) {
      notify(player, "You cannot link things to anywhere from afar.");
      return 0;
    }
  } if(room == NOTHING)  /* Unset a dropto on a room only. */
    return 1;

  if(room == HOME) {
    if(Typeof(thing) != TYPE_EXIT && Typeof(thing) != TYPE_ROOM) {
      notify(player, tprintf("%s need a firm reference point to link to.",
             Typeof(thing) == TYPE_ZONE ? "Zones":"Objects"));
      return 0;
    }
  } else {
    if((Typeof(thing) == TYPE_THING && (Typeof(room) == TYPE_EXIT ||
       Typeof(room) == TYPE_ZONE)) || (Typeof(thing) != TYPE_THING &&
       Typeof(room) != TYPE_ROOM)) {
      notify(player, tprintf(
             "Object type %s may not be linked to object type %s.",
             typenames[Typeof(thing)], typenames[Typeof(room)]));
      return 0;
    }

    /* check standard flags */
    if(!controls(player, room, POW_MODIFY) &&
      ((Typeof(thing) != TYPE_PLAYER && !(db[room].flags & ROOM_LINK_OK)) ||
       (Typeof(thing) == TYPE_PLAYER && !(db[room].flags & ROOM_ABODE)))) {
      notify(player, "Permission denied.");
      return 0;
    }

    /* check room lock */
    if(Typeof(thing) != TYPE_ZONE && !could_doit(player, room, A_LLINK)) {
      notify(player, "Permission denied.");
      return 0;
    }

    /* check restricted zone */
    if(Typeof(thing) == TYPE_EXIT) {
      for(ptr=db[mainroom(thing)].zone;ptr && *ptr != NOTHING;ptr++)
        if(!controls(player, *ptr, POW_MODIFY) &&
           !inlist(db[mainroom(room)].zone, *ptr)) {
          notify(player, "Cannot link to a restricted zone.");
          return 0;
        }
      for(ptr=db[mainroom(room)].zone;ptr && *ptr != NOTHING;ptr++)
        if(!controls(player, *ptr, POW_MODIFY) &&
           !inlist(db[mainroom(thing)].zone, *ptr)) {
          notify(player, "Cannot link from a restricted zone.");
          return 0;
        }
    }
  }

  if(Typeof(thing) == TYPE_EXIT) {
    /* linking exits is just like opening an exit. check builder permissions */
    if(RESTRICT_BUILD && !Builder(player)) {
      notify(player, "Only authorized builders may link exits.");
      return 0;
    }

    /* linking exits requires a small fee */
    if(!power(player, POW_FREE) && Typeof(thing) == TYPE_EXIT &&
       db[thing].link == NOTHING && !charge_money(player, LINK_COST, 0)) {
      notify(player, tprintf("It costs %d %s to link an exit.",
             LINK_COST, (LINK_COST==1)?MONEY_SINGULAR:MONEY_PLURAL));
      return 0;
    }
  }

  return 1;
}

/* use this to create a room */
void do_dig(dbref player, char *name, char *argc, char **argv)
{
  char buf[strlen(name)+1], *s;
  int cost, room, *ptr;
  
  /* check building permissions */
  if(RESTRICT_BUILD && !Builder(player)) {
    notify(player, "Only authorized builders may dig rooms.");
    return;
  } if(Guest(player)) {
    notify(player, "Guests cannot dig rooms.");
    return;
  } if(!power(player, POW_SECURITY) && (!controls(player, db[player].location,
       POW_MODIFY) || Typeof(db[player].location) != TYPE_ROOM)) {
    notify(player, "You can't @dig from there.");
    return;
  } if(!*name) {
    notify(player, "You must specify a name for the room.");
    return;
  } if(!ok_name(name)) {
    notify(player, "That is an unacceptable room name.");
    return;
  }

  /* Check cost of exits */
  cost=ROOM_COST + ((argv[1] && *argv[1])?(EXIT_COST+LINK_COST):0) +
                   ((argv[2] && *argv[2])?(EXIT_COST+LINK_COST):0);
  if(!charge_money(player, cost, 2) || !charge_money(player, ROOM_COST, 1))
    return;

  if((room=new_object()) == NOTHING) {
    notify(player, "The database has reached the maximum number of objects. "
           "Please @destroy something first.");
    return;
  }

  /* Initialize everything */
  strcpy(buf, name);
  if((s=strchr(buf, ';'))) {
    *s=0;
    while(s > buf && *(s-1) == ' ')
      *--s=0;
  }
  WPTR(db[room].name,*buf?buf:"Room");
  db[room].owner = db[player].owner;
  db[room].flags = TYPE_ROOM;
  db[room].location = room;
  db[room].create_time = now;
  atr_add(room, A_FLOOR, atr_get(db[player].location, A_FLOOR));
  atr_add(room, A_COLOR, atr_get(db[player].location, A_COLOR));
  atr_add(room, A_WEATHER, atr_get(db[player].location, A_WEATHER));
  notify(player, tprintf("%s created with room number #%d.", name, room));

  /* copy zones */
  for(ptr=db[db[player].location].zone;ptr && *ptr != NOTHING;ptr++)
    PUSH_L(db[room].zone, *ptr);

  /* link proper exits if specified and if enough funds are present */
  if(argv[1] && *argv[1] &&
     open_exit(player, argv[1], db[player].location, room))
    if(argv[2] && *argv[2])
      open_exit(player, argv[2], room, db[player].location);
}

/* use this to create/link an exit */
void do_open(dbref player, char *arg1, char *arg2)
{
  dbref dest=NOTHING;

  if(*arg2 && (dest=match_thing(player, arg2, MATCH_EVERYTHING|MATCH_HOME)) ==
     NOTHING)
    return;

  open_exit(player, arg1, db[player].location, dest);
}

/* creates exit in loc. if destination != nothing, links it for a fee */
int open_exit(dbref player, char *name, dbref loc, dbref dest)
{
  char buf[strlen(name)+1], *p, *r, *s;
  dbref exit;
  
  if(RESTRICT_BUILD && !Builder(player)) {
    notify(player, "Only authorized builders may create exits.");
    return 0;
  } if(Guest(player)) {
    notify(player, "Guests cannot open exits.");
    return 0;
  } if(Invalid(loc) || Typeof(loc) != TYPE_ROOM) {
    notify(player, tprintf("An exit is not allowed from %s.", db[loc].name));
    return 0;
  } if(!*name) {
    notify(player, "You must specify a name for the exit.");
    return 0;
  } if(!controls(player, loc, POW_MODIFY)) {
    notify(player, "Permission denied.");
    return 0;
  } if(!ok_name(name)) {
    notify(player, "That is an unacceptable exit name.");
    return 0;
  } if(!charge_money(player, EXIT_COST, 1))
    return 0;

  /* determine exit name */
  strcpy(buf, name);
  s=buf;
  r=(char *)parse_up(&s, ';');
  if((p=strchr(r, '\0')))
    while(p > r && *(p-1) == ' ')
      *--p=0;

  if((exit=new_object()) == NOTHING) {
    notify(player, "The database has reached the maximum number of objects. "
           "Please @destroy something first.");
    return 0;
  }

  /* initialize everything */
  WPTR(db[exit].name,*r?r:"Exit");
  atr_add(exit, A_ALIAS, s);
  db[exit].owner = db[player].owner;
  db[exit].flags = TYPE_EXIT;
  db[exit].plane = -1;
  db[exit].create_time = now;
    
  /* add exit to roomlist */
  PUSH(exit, db[loc].exits);
  db[exit].location = loc;
  db[exit].link = NOTHING;

  /* check to see if we should link to destination */
  if(dest != NOTHING && can_link(player, exit, dest)) {
    db[exit].link = dest;
    notify(player, tprintf("%s opened; linked to %s.", db[exit].name,
           unparse_object(player, dest)));
    return 1;
  }

  /* exit unsuccessfully set a destination */
  notify(player, tprintf("%s opened.", db[exit].name));
  return 0;
}

/* use this to create an object */
void do_create(dbref player, char *name)
{
  char buf[strlen(name)+1], *s;
  dbref thing;
  
  /* check building permissions */
  if(Typeof(player) == TYPE_ZONE) {
    notify(player, "Zones cannot create objects.");
    return;
  } if(RESTRICT_BUILD && !Builder(player)) {
    notify(player, "Only authorized builders may create objects.");
    return;
  } if(Guest(player)) {
    notify(player, "Guests cannot create objects.");
    return;
  } if(!*name) {
    notify(player, "You must specify a name for the object.");
    return;
  } if(!ok_name(name)) {
    notify(player, "That is an unacceptable object name.");
    return;
  } if(!charge_money(player, THING_COST, 1))
    return;

  if((thing=new_object()) == NOTHING) {
    notify(player, "The database has reached the maximum number of objects. "
           "Please @destroy something first.");
    return;
  }

  /* initialize everything */
  strcpy(buf, name);
  if((s=strchr(buf, ';'))) {
    *s=0;
    while(s > buf && *(s-1) == ' ')
      *--s=0;
  }
  WPTR(db[thing].name,*buf?buf:"Thing");
  db[thing].owner = db[player].owner;
  db[thing].flags = TYPE_THING;
  db[thing].plane = (Typeof(player)==TYPE_ROOM || Typeof(player)==TYPE_EXIT)?
                    db[player].plane:-1;
  db[thing].create_time = now;

  /* move object to player's inventory */
  db[thing].location=(Typeof(player) == TYPE_EXIT)?db[player].location:player;
  PUSH(thing, db[db[thing].location].contents);

  /* set a default link */
  if(Typeof(db[player].location) == TYPE_ROOM &&
     controls(player, db[player].location, POW_MODIFY))
    db[thing].link = db[player].location;
  else
    db[thing].link = player;

  notify(player, tprintf("%s created.", unparse_object(player, thing)));
}

void do_clone(dbref player, char *arg1, char *arg2)
{
  char buf[strlen(arg2)+1], *s;
  dbref thing, clone, *ptr;
  
  if(clone_depth) {
    notify(player, "Recursive @clone detected.");
    return;
  }

  /* check building permissions */
  if(RESTRICT_BUILD && !Builder(player)) {
    notify(player, "Only authorized builders may clone objects.");
    return;
  } if(Guest(player)) {
    notify(player, "Guests cannot clone objects.");
    return;
  } if((thing=match_thing(player, arg1, MATCH_EVERYTHING)) == NOTHING)
    return;
  if(!controls(player, thing, POW_MODIFY)) {
    notify(player, "Permission denied.");
    return;
  } if(Typeof(thing) != TYPE_THING) {
    notify(player, "You can only clone objects.");
    return;
  } if(db[thing].bitmaplen && !power(player, POW_COMBAT)) {
    notify(player, "Objects set with a fight status may not be cloned.");
    return;
  } if(!charge_money(player, THING_COST, 1))
    return;

  /* check for valid newname */
  if(*arg2 && !ok_name(arg2)) {
    notify(player, "That is an unacceptable object name.");
    return;
  }

  if((clone=new_object()) == NOTHING) {
    notify(player, "The database has reached the maximum number of objects. "
           "Please @destroy something first.");
    return;
  }

  if(*arg2) {
    strcpy(buf, arg2);
    if((s=strchr(buf, ';'))) {
      *s=0;
      while(s > buf && *(s-1) == ' ')
        *--s=0;
    }
  } else
    *buf=0;
  WPTR(db[clone].name, *buf?buf:db[thing].name);
  db[clone].owner = db[player].owner;
  db[clone].create_time = now;
  db[clone].flags = db[thing].flags;
  db[clone].plane = db[thing].plane;

  /* fix up parents/children/relay list */
  for(ptr=db[thing].parents;ptr && *ptr != NOTHING;ptr++) {
    if(db[*ptr].flags & BEARING || controls(player, *ptr, POW_MODIFY)) {
      PUSH_L(db[clone].parents, *ptr);
      PUSH_L(db[*ptr].children, clone);
    }
  }
  for(ptr=db[thing].relay;ptr && *ptr != NOTHING;ptr++)
    PUSH_L(db[clone].relay, *ptr);

  atr_cpy(clone, thing);
  notify(player, tprintf("%s cloned with number #%d.",
         unparse_object(player, thing), clone));

  /* push object into the room's (or player's) contents */
  db[clone].location=(Typeof(player) == TYPE_EXIT) ? db[player].location:
     (Typeof(player) == TYPE_ZONE) ? db[thing].location:player;
  PUSH(clone, db[db[clone].location].contents);

  /* set a default link */
  if(Typeof(db[player].location) == TYPE_ROOM && controls(player,
     db[player].location, POW_MODIFY))
    db[clone].link = db[player].location;
  else
    db[clone].link = player;

  /* trigger enemy startup */
  if(*(s=atr_get(clone, A_ACLONE))) {
    clone_depth=1;
    immediate_que(clone, s, clone);
    clone_depth=0;
  }

  /* active @aclone on new object */
  did_it(player, clone, 0, NULL, A_OCLONE, NULL, 0);
}

/* creates a zone object */
void do_zone(dbref player, char *name, char *link)
{
  char buf[strlen(name)+1], *s;
  dbref zone, room;
  
  /* check building permissions */
  if(RESTRICT_BUILD && !Builder(player)) {
    notify(player, "Only authorized builders may create zones.");
    return;
  } if(Guest(player)) {
    notify(player, "Guests cannot create zones.");
    return;
  } if(!*name) {
    notify(player, "You must specify a name for the zone.");
    return;
  } if(!ok_name(name)) {
    notify(player, "That is an unacceptable zone name.");
    return;
  } if(!charge_money(player, ZONE_COST, 1))
    return;

  if((zone=new_object()) == NOTHING) {
    notify(player, "The database has reached the maximum number of objects. "
           "Please @destroy something first.");
    return;
  }

  /* initialize everything */
  strcpy(buf, name);
  if((s=strchr(buf, ';'))) {
    *s=0;
    while(s > buf && *(s-1) == ' ')
      *--s=0;
  }
  WPTR(db[zone].name,*buf?buf:"Zone");
  db[zone].owner = db[player].owner;
  db[zone].location = zone;
  db[zone].flags = TYPE_ZONE;
  db[zone].create_time = now;

  notify(player, tprintf("Zone %s created.", unparse_object(player, zone)));

  /* add origin toom if specified */
  if(*link) {
    if((room=match_thing(player, link, MATCH_EVERYTHING)) == NOTHING)
      return;
    if(!can_link(player, zone, room))
      return;

    db[zone].link = room;
    notify(player, tprintf("Origin Room set to %s.", unparse_object(player,
           room)));
  }
}
