/* db/attrib.c */
/* Database attribute maintainence procedures */

#include "externs.h"

ATTR attr[]={
#define DOATTR(num, var, name, flags) {name, NOTHING, num, flags},
#define VREGS
#include "attrib.h"
#undef VREGS
#undef DOATTR
{NULL, 0, 0, 0}};

int db_attrs;  /* Keeps count of the number of attributes stored */


/* lists all attributes to player */
void list_attributes(dbref player, char *name)
{
  int a, b;

  if(*name) {
    b=atoi(name);
    for(a=0;attr[a].num;a++)
      if(attr[a].num == b || !strcasecmp(attr[a].name, name)) {
        notify(player, tprintf("%4d: %-15.15s %s", attr[a].num,
               attr[a].name, unparse_attrflags(attr[a].flags)));
        return;
      }
    notify(player, tprintf("No such built-in attribute '%s'.", name));
    return;
  }

  notify(player, "Numerical list of attributes and their hardcoded options:");
  notify(player, "");
  for(a=0;attr[a].num;a++)
    notify(player, tprintf("%4d: %-15.15s %s", attr[a].num,
           attr[a].name, unparse_attrflags(attr[a].flags)));
}


/* pointer-returning functions */

/* returns the pointer to a builtin attribute given its name */
ATTR *builtin_atr_str(char *str)
{
  static hash *attrhash;

  if(!attrhash)
    attrhash=make_hashtab(attr, sizeof(ATTR));
  return lookup_hash(attrhash, str);
}

/* given an attribute name <str>, looks up an attr set on <obj> */
ATTR *atr_str(dbref obj, char *str)
{
  dbref *ptr;
  ATRDEF *k;
  ATTR *a;

  /* search for defined attribute on <obj> or <obj>'s parents (recursive) */
  for(k=db[obj].atrdefs;k;k=k->next)
    if(!strcasecmp(k->atr->name, str))
      return k->atr;
  for(ptr=db[obj].parents;ptr && *ptr != NOTHING;ptr++)
    if((a=atr_str(*ptr, str)))
      return a;

  return builtin_atr_str(str);
}

/* get the attribute definition for #obj.num */
ATTR *atr_num(dbref obj, unsigned char num)
{
  static ATTR *attrs[256];
  static int init=0;
  ATRDEF *k;
  int a;

  /* Built-in Attribute */
  if(obj == NOTHING) {
    /* allocate index on first run */
    if(!init) {
      for(a=0;attr[a].num;a++)
        attrs[attr[a].num]=&attr[a];
      init=1;
    }
    return attrs[num];
  }

  /* User-defined Attribute */
  for(k=db[obj].atrdefs;k;k=k->next)
    if(k->atr->num == num)
      return k->atr;
  return NULL;
}


/* routines to handle object attribute lists */

/* adds defined attr <num> on <obj> to <thing> */
void atr_add_obj(dbref thing, dbref obj, unsigned char num, char *str)
{
  ALIST *ptr, *last=NULL;

  db[thing].mod_time=now;
  str=compress(str);

  for(ptr=db[thing].attrs;ptr;ptr=ptr->next) {
    if(ptr->obj == obj && ptr->num == num) {
      /* erase attribute? */
      if(!*str) {
        if(last)
          last->next=ptr->next;
        else
          db[thing].attrs=ptr->next;
        free(ptr);
        db_attrs--;
        return;
      }

      /* otherwise set new value on attribute */
      ptr=realloc(ptr, sizeof(ALIST)+strlen(str)+1);
      if(last)
        last->next=ptr;
      else
        db[thing].attrs=ptr;

      strcpy(ptr->text, str);
      return;
    }
    last=ptr;
  }

  if(!*str)  /* nothing to clear? */
    return;

  /* add attribute */
  db_attrs++;
  ptr=malloc(sizeof(ALIST)+strlen(str)+1);
  ptr->next=NULL;
  ptr->obj=obj;
  ptr->num=num;
  strcpy(ptr->text, str);
  if(last)
    last->next=ptr;
  else
    db[thing].attrs=ptr;
}

/* gets defined attr <num> on <obj> from <thing> */
char *atr_get_obj(dbref thing, dbref obj, unsigned char num)
{
  dbref *stack[16];
  int index=-1;
  char *s;

  ALIST *ptr;
  ATTR *a;

  while(1) {
    /* Scan attribute list for matches */
    for(ptr=db[thing].attrs;ptr;ptr=ptr->next)
      if(ptr->num == num && ptr->obj == obj)
        return uncompress(ptr->text);

    /* Check (only once) that the attribute is Inheritable */
    if(index == -1 && (!(a=atr_num(obj, num)) || !(a->flags & INH)))
      break;

    /* Advance stack pointer */
    if(index > -1 && *(++stack[index]) == NOTHING)
      index--;

    /* Add parents to the stack */
    if(thing != obj && db[thing].parents)
      stack[++index]=db[thing].parents;

    if(index == -1)
      break;
    if(index == 15) {
      while(*stack[index] != NOTHING) {
        if(*(s=atr_get_obj(*stack[index], obj, num)))
          return s;
        ++stack[index];
      }
      index--;
    }

    thing=*stack[index];
  }

  return "";
}

// When destroying <obj> or delparenting <thing> from <obj>, remove all
// attributes that have references to <obj> on its children (nonrecursively) */
//
void atr_clean(dbref thing, dbref obj)
{
  dbref **stack=NULL;
  int size=0, cur=0;

  ALIST *ptr, *last=NULL, *next;

  while(1) {
    /* Wipe all attributes on <thing> belonging to <obj> */
    for(ptr=db[thing].attrs;ptr;ptr=next) {
      next=ptr->next;
      if(ptr->obj == obj) {
        if(last)
          last->next=next;
        else
          db[thing].attrs=next;
        free(ptr);
        db_attrs--;
      } else
        last=ptr;
    }

    /* Advance stack pointer */
    if(cur && *(++stack[cur-1]) == NOTHING)
      cur--;

    /* Add child members to the stack */
    if(db[thing].children) {
      if(++cur > size)
        stack=realloc(stack, (size+=4)*sizeof(dbref *));
      stack[cur-1]=db[thing].children;
    }

    if(!cur)
      break;
    thing=*stack[cur-1];
  }

  free(stack);
}

/* copy attributes from source to dest */
void atr_cpy(dbref dest, dbref source)
{                 
  ALIST *ptr, *new, *last=NULL;
  ATRDEF *k, *newk, *lastk=NULL;
  
  /* copy attribute definitions */
  for(k=db[source].atrdefs;k;k=k->next) {
    newk=malloc(sizeof(ATRDEF));
    newk->next=NULL;
    newk->atr=malloc(sizeof(ATTR));
    newk->atr->obj=dest;
    newk->atr->num=k->atr->num;
    newk->atr->flags=k->atr->flags;
    SPTR(newk->atr->name, k->atr->name);
    if(lastk)
      lastk->next=newk;
    else
      db[dest].atrdefs=newk;
    lastk=newk;
  }

  /* copy attribute data */
  for(ptr=db[source].attrs;ptr;ptr=ptr->next) {
    if(ptr->obj != NOTHING && source != ptr->obj && !is_a(dest, ptr->obj))
      continue;
    db_attrs++;
    new=malloc(sizeof(ALIST)+strlen(ptr->text)+1);
    new->next=NULL;
    new->obj=(ptr->obj == source)?dest:ptr->obj;
    new->num=ptr->num;
    strcpy(new->text, ptr->text);
    if(last)
      last->next=new;
    else
      db[dest].attrs=new;
    last=new;
  }
}

// Returns a allocated list of all attributes (local or inherited) that
// are defined on an object. Results MUST be freed when done using them.
//
struct all_atr_list *all_attributes(dbref thing)
{
  dbref **stack=NULL;
  int size=0, dep=0;

  struct all_atr_list *list=NULL;
  int i, base=0, count=0;
  ATTR *atr;
  ALIST *a;

  while(1) {
    for(a=db[thing].attrs;a;a=a->next) {
      if(!(atr=atr_num(a->obj, a->num)) || (dep && !(atr->flags & INH)))
        continue;

      /* Check to see if there's already something with this type */
      for(i=0;i < count && list[i].type != atr;i++);
      if(i < count)
        continue;

      /* Increase the number of attributes returned by 10 (+1 for EOL) */
      if(!(count++ % 10))
        list=realloc(list, (count+10)*sizeof(struct all_atr_list));
      list[i].type=atr;
      list[i].value=a->text;
      list[i].numinherit=dep+base;
    }

    /* Advance stack pointer */
    if(dep && *(++stack[dep-1]) == NOTHING) {
      base++;
      dep--;
    }

    /* Add parents to the stack */
    if(db[thing].parents) {
      if(++dep > size)
        stack=realloc(stack, (size+=4)*sizeof(dbref *));
      stack[dep-1]=db[thing].parents;
    }

    if(!dep)
      break;
    thing=*stack[dep-1];
  }

  /* Mark end of list by setting all fields to NULL */
  if(list)
    memset(&list[count], 0, sizeof(struct all_atr_list));

  free(stack);
  return list;
}
