/* mare/runtime.c */
/* runtime control and database invocation commands */

#include "externs.h"
#include <unistd.h>
#include <sys/resource.h>

int epoch_fail;
int halt_dump;


void do_dump(dbref player)
{
  if(db_readonly) {
    notify(player, tprintf("Cannot save the database. "
           "%s was loaded in read-only mode.", MUD_NAME));
    return;
  }
  if(halt_dump) {
    notify(player, "Sorry, a database save is already in effect.");
    return;
  }
  notify(player, "Dumping...");
  fork_and_dump();
  dump_interval=0;
}

void do_shutdown(dbref player)
{
  log_main(tprintf("SHUTDOWN: %s(#%d)", db[player].name, player));
  shutdown_flag=1;
}

void do_reboot(dbref player, char *options)
{
  const char *flags[3]={"Restart", "Quiet", "Validate"};
  char tmp[32], *r, *s=options;
  int i;

  boot_flags=0;
  *tmp='\0';

  while((r=parse_up(&s,' '))) {
    for(i=0;i<3;i++)
      if(string_prefix(flags[i], r))
        break;
    if(i == 3) {
      notify(player, tprintf("No such bootflag '%s'.", r));
      return;
    }

    boot_flags |= 1 << i;
    sprintf(tmp+strlen(tmp), "%s%s", *tmp?", ":" - ", flags[i]);
  }

  log_main(tprintf("REBOOT: %s(#%d)%s", db[player].name, player, tmp));
  shutdown_flag=2;
}

/* for use with background database dumping only */
static void close_fds()
{
  DESC *d;

  for(d=descriptor_list;d;d=d->next)
    if(!(d->flags & C_REMOTE))
      close(d->descriptor);
}

static int dump_database_internal()
{
  char tmpfile[256];
  FILE *f;
  
  sprintf(tmpfile, "%s.#%d", db_file, nstat[NS_DBSAVE]-1);
  unlink(tmpfile);
  sprintf(tmpfile, "%s.#%d", db_file, nstat[NS_DBSAVE]);

  if(!(f=fopen(tmpfile, "wb"))) {
    perror(tmpfile);
    return 0;
  }

  if(!db_write(f, tmpfile)) {
    fclose(f);
    unlink(tmpfile);
    return 0;
  }

  /* send all data to disk immediately */
  fflush(f);
  fsync(fileno(f));
  fclose(f);

  if(rename(tmpfile, db_file)) {
    perror(tmpfile);
    return 0;
  }

  return 1;  /* save successful */
}

void dump_database()
{
  if(db_top < 2 || Typeof(GOD) != TYPE_PLAYER)   /* security check */
    return;

  nstat[NS_DBSAVE]++;
  log_main(tprintf("Database Save: %s.#%d", db_file, nstat[NS_DBSAVE]));
  if(!dump_database_internal()) {
    if(shutdown_flag)
      log_main(tprintf("%s Dump Error. Database Lost!!", MUD_NAME));
    else
      log_main(tprintf("%s Dump Failed.", MUD_NAME));
  } else
    log_main(tprintf("%s Dumping Done.", MUD_NAME));
}

void fork_and_dump()
{
  int child;
  
  if(db_top < 2 || Typeof(GOD) != TYPE_PLAYER)   /* security check */
    return;

  nstat[NS_DBSAVE]++;
  log_main(tprintf("Database Checkpointing: %s.#%d", db_file,
           nstat[NS_DBSAVE]));
  
  if(DB_FORK)
    child=fork();
  else {
    /* warn people that the game will stop for a few seconds.. */
    broadcast(tprintf("%s Autosave...!\n", MUD_NAME));
    child=0;
  }
  
  if(child < 0) {
    perror("dbsave: fork()");
    no_dbdump(1);
    return;
  }

  if(child > 0) {  /* parent process */
    halt_dump=1;
    return;
  }

  /* child process or non-forking save: */

  if(DB_FORK) {
    /* unset virtual timer and lower background process priority */
    alarm(0);

    /* set all signals to default handlers */
    signal(SIGSEGV, SIG_DFL);
    signal(SIGBUS, SIG_DFL);
    signal(SIGILL, SIG_DFL);
    signal(SIGFPE, SIG_DFL);

    /* close remaining files */
    close_fds();

#ifdef PRIO_PROCESS
    /* enter low background priority */
    setpriority(PRIO_PROCESS, 0, getpriority(PRIO_PROCESS, 0)+4);
#endif
  }

  if(!dump_database_internal()) {
    if(DB_FORK)
      exit(1);  /* terminate background process */
    no_dbdump(0);
  }
  if(DB_FORK)
    exit(0);  /* forked save successful */

  epoch_fail=0;
  if(fsize(db_file) >= 0)
    log_main(tprintf("%s Checkpoint Done. Filesize: %ld bytes", MUD_NAME,
             fsize(db_file)));
}

void no_dbdump(int background)
{
  broadcast(tprintf("\aGame Driver: Database save #%d failed. "
            "Please take appropriate precautions.\n", nstat[NS_DBSAVE]));
  epoch_fail += background;

  if(epoch_fail > 9) {
    broadcast("\aGame Driver: Manually Saving. "
              "Response may be slow for a few minutes.\n");
    epoch_fail=0;
    dump_database();
    dump_interval=0;
  } else
    dump_interval=DUMP_INTERVAL-300;  /* 5-minute delay */
}

void do_crash(dbref player)
{
  DESC *d;

  /* notify online */
  log_main(tprintf("%s(#%d) executes @crash. G'nite World!",
           db[player].name, player));
  for(d=descriptor_list;d;d=d->next) {
    output2(d, "Crashing...!\n");
    process_output(d);
  }

  /* terminated. */
  exit(2);
}


/* kick-start game execution */
void start_game()
{
  char *s;
  dbref thing;
  int a;

  /* enable the game driver for command processing */
  db[GAME_DRIVER].flags &= ~HAVEN;

  /* execute @startup attribute */
  for(thing=0;thing<db_top;thing++)
    if(!(db[thing].flags & (GOING|HAVEN)) && 
       db[thing].create_time < now && *(s=atr_get(thing, A_STARTUP))) {
      immediate_que(thing, s, thing);
      for(a=0;a<10;a++)
        *wptr[a]='\0';
    }
}

/* initialize database configuration upon startup */
void init_config()
{
  /* initialize gametime on new mare database */
  if(!BEGINTIME) {
    log_main(tprintf("MARE Startdate - January 1, 0001: %s", strtime(now)));
    log_main("");
    BEGINTIME=now;
    FIRST_YEAR=1;
  }

  init_dbstat();	/* update database accounting information */
}

/* boot up into "single-user mode" and read database file */
void test_database()
{
  FILE *f;

  /* initialize a limited system environment */
  init_compress();
  disable_startups=1;

  /* open database file */
  if(!(f=fopen(db_file, "rb"))) {
    perror(db_file);
    return;
  }

  db_read_file=f;
  log_main(tprintf("Reading file '%s'...", db_file));

  log_main(tprintf(" Date: %s", strtime(fdate(db_file))));
  log_main(tprintf(" Size: %ld bytes", fsize(db_file)));

  /* read entire db file */
  while(!load_database());

  db_print(stdout);
}

void db_print(FILE *f)
{
  ATTR *atr;
  ALIST *a;
  ATRDEF *k;
  int i;

  fprintf(f, "\n***Begin Dump***\n\n");
  for(i=0;i<db_top;i++) {
    fprintf(f, "%s(#%d%s)\n", db[i].name, i, unparse_flags(GOD, i, 0));
    fprintf(f, "Location: #%-10d Owner: %s\n",
            db[i].location, db[db[i].owner].name);
    fprintf(f, "Link:     #%-10d Type:  %s\n",
            db[i].link, typenames[Typeof(i)]);
    fprintf(f, "Bytes:    %-10d  Flags: %s\n",
            mem_usage(i), unparse_flags(GOD, i, 1));

    if(db[i].create_time)
      fprintf(f, "Created: %s\n", mktm(db[i].create_time, GOD));
    if(db[i].mod_time)
      fprintf(f, "Modified: %s\n", mktm(db[i].mod_time, GOD));
    for(k=db[i].atrdefs; k; k=k->next)
      fprintf(f, "  %s%s%s%s\n",k->atr->name,(k->atr->flags & AF_FUNC)?"()":"",
              k->atr->flags?": ":"", unparse_attrflags(k->atr->flags));

    for(a=db[i].attrs;a;a=a->next)
      if((atr=atr_num(a->obj, a->num)))
        fprintf(f, "%s:%s\n", atr->name, uncompress(a->text));
    fprintf(f, "\n");
  }
  fprintf(f, "***End of Dump***\n");
}
