/* io/net.c */
/* Main server module, handles all network connections. */

#include "externs.h"
#include <errno.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <sys/utsname.h>
#include <sys/socket.h>
#include <arpa/telnet.h>
#include <netdb.h>

char *db_file = "db/mdb";
char **mainarg;

int nstat[NET_STATS];
int shutdown_flag;
int game_online;
int alarm_count;
int sig_caught;
int save_flag;
int badforksave;
int resolv=-1;
int xcpu;
int dberrcheck=1;
int disable_startups;
int boot_flags;
int mare_test_db;
int db_forceload;
int db_readonly;

long open_files=-1;
long mud_up_time;
long last_reboot_time;
long now;
quad time_slice;

char system_info[64];
char system_version[64];
char system_model[16];
int  system_ncpu;
int  system_cpu_mhz;

char global_buff[16384];
char log_text[8192];
dbref log_dbref=NOTHING;
char *wptr[10];

DESC *descriptor_list;

int tinyport=DEFAULT_PORT;


/* startup & initialize everything */
int main(int argc, char *argv[])
{
  /* process command line flags and arguments */
  init_args(argc, argv);

  /* initialize input & output */
  init_io();

  /* initialize game structure */
  init_game();
  
  /* start game driver */
  serverport();

  return 0;
}

#ifdef linux

// Gather CPU information from /proc/cpuinfo (Linux only).
// Returns the number of CPUs detected plus additional information in
// system_ncpu, system_model, and system_cpu_mhz.
//
static void get_cpuinfo()
{
  FILE *f;
  char buf[256], *s, *t;
  int mhz=0;

  if(!(f=fopen("/proc/cpuinfo", "r")))
    return;

  while(ftext(buf, 256, f)) {
    if(!(s=strchr(buf, ':')))
      continue;
    for(t=s-1;t > buf && isspace(*t);t--);
    *(t+1)='\0';
    for(s++;isspace(*s);s++);
    if(!*s)
      continue;

#ifdef __alpha__
    if(!strcasecmp(buf, "cpus detected"))
      system_ncpu=atoi(s);
    else if(!strcasecmp(buf, "cycle frequency [Hz]"))
      mhz=atoi(s)/1000000;
    else if(!strcasecmp(buf, "cpu model")) {
      strcpy(system_model, "alpha-");
      strmaxcpy(system_model+6, s, 10);
    }
#else
    if(!strcasecmp(buf, "processor"))
      system_ncpu++;
    else if(!strcasecmp(buf, "cpu MHz") || !strcasecmp(buf, "Clocking") ||
            !strcasecmp(buf, "clock"))
      mhz=atoi(s);
#endif
  }
  fclose(f);

  /* Round MHz to the nearest multipler */
  if((mhz+69) % 100 < 7)
    mhz=mhz/100*100+33;
  else if((mhz+37) % 100 < 6)
    mhz=mhz/100*100+66;
  else if(mhz % 5 < 3)
    mhz-=mhz % 5;
  else
    mhz+=5-(mhz % 5);
  if(mhz == 666)
    mhz++;
  system_cpu_mhz=mhz;

  /* Make architecture model name lowercase */
  for(s=system_model;*s;s++)
    *s=tolower(*s);

  if(!system_ncpu)
    system_ncpu=1;
}
#endif  /* linux */

void init_args(int argc, char *argv[])
{
  struct utsname sys;
  char option, *cmd=argv[0], getinput=0;
  int a;

  mainarg=argv;

  while(argc > 1) {
    --argc; ++argv;
    if(getinput) {
      db_file=*argv;
      getinput=0;
      continue;
    }

    if(**argv == '-') {
      for(a=1;(*argv)[a];a++) {
        switch((option=(*argv)[a])) {
        case 'd': /* set default database input/output file */
          getinput=option;
          break;
        case 'f': /* force loading the database without combat structures */
          db_forceload=1;
          break;
        case 'r': /* make database file read-only */
          db_readonly=1;
          break;
        case 's': /* disable startups */
          disable_startups=1;
          break;
        case 't': /* test load database and dump data to stdout */
          mare_test_db=1;
          getinput=option;
          break;
        case 'v': /* display version and exit */
          fprintf(stdout, "%s: Version Reference: %s\n", cmd, mud_version);
          exit(0);
        default:
          fprintf(stderr, "%s: illegal option -- %c\n", cmd, option);
          fprintf(stderr, "Usage: %s [-frsv] [-d|-t file] [port]\n", cmd);
          exit(0);
        }
      }
      continue;
    }

    /* change port number? */
    tinyport=atoi(*argv);
    if(tinyport < 1 || tinyport > 65535) {
      fprintf(stderr, "%s: Invalid port number %d\n", cmd, tinyport);
      exit(0);
    } if(tinyport < 1024 && geteuid()) {
      fprintf(stderr, "%s: Only root can run on port %d\n", cmd, tinyport);
      exit(0);
    }
    break;
  }

  /* validate multiword options */
  if(getinput) {
    fprintf(stderr, "%s: Option '%c' takes an argument.\n", cmd, getinput);
    exit(0);
  }

  /* extract system info */
  uname(&sys);
  sprintf(system_info, "%s %s %s (%s)", sys.machine, sys.sysname,
          sys.release, sys.nodename);
  sprintf(system_version, "%s %s", sys.sysname, sys.release);
  strmaxcpy(system_model, sys.machine, 16);

#ifdef linux
  get_cpuinfo();  /* Determine CPU Model and MHz */
#endif

  /* determine max number of open files allowed */
#ifdef __CYGWIN__
  open_files=-1;
#elif defined(RLIMIT_NOFILE)
  {
    struct rlimit rlim;
    getrlimit(RLIMIT_NOFILE, &rlim);
    open_files=rlim.rlim_cur;
  }
#else
  open_files=sysconf(_SC_OPEN_MAX);
#endif

  if((unsigned int)open_files > 0x7ffffffe)
    open_files=-1;
}

void init_io()
{
  char *ptr;
  int a;

  now=time(NULL);

  /* reset i/o statistics */
  for(a=0;a<NET_STATS;a++)
    nstat[a] = 0;

  /* initialize environment pointers */
  ptr=malloc(8192*10);
  for(a=0;a<10;a++)
    *(wptr[a]=ptr+8192*a)='\0';

  /* test database only? (-t option) */
  if(mare_test_db) {
    test_database();
    exit(0);
  }

  /* ignore all pipe and floating point exception signals */
  signal(SIGPIPE, SIG_IGN);
  signal(SIGUSR2, SIG_IGN);
  signal(SIGFPE, error_handler);  // does not ignore properly!

  /* alarm routines */
  signal(SIGALRM, alarm_handler);

  /* database save signal */
  signal(SIGCLD, reaper);

  /* trap normal signals */
  signal(SIGQUIT, sig_handler);
  signal(SIGTERM, sig_handler);
  signal(SIGUSR1, sig_handler);
  signal(SIGHUP, sig_handler);
  signal(SIGINT, sig_handler);

  /* trap error signals */
  signal(SIGILL, error_handler);
  signal(SIGBUS, error_handler);
  signal(SIGSEGV, error_handler);

  if(open_sockets())  /* reboot mode */
    return;

  /* All error output is logged to logs/error instead,  and TinyMare's */
  /* Main Output is logged to logs/main, with time prefixing each line */

  fclose(stdin);
  fclose(stdout);
  fclose(stderr);

  mud_up_time=now;
  nstat[NS_EXEC]++;
}

void init_game()
{
  FILE *f;

  init_compress();

  /* initialize command counters */
  cmdav[0]=nstat[NS_NPCMDS];
  cmdav[1]=nstat[NS_NQCMDS];

  /* initialize random number generator */
  srandom(now);

  last_reboot_time=now;

  log_main("---");
  log_main(tprintf("Running %s - %s", mud_version, system_info));
  log_main(tprintf("Starting Program Execution #%d", nstat[NS_EXEC]));
  log_main("");

  /* attempt to open all login sockets */
  open_ports();

  /* open database file for reading */
  if((f=fopen(db_file, "rb")) == NULL) {
    if(mare_test_db || db_readonly || errno != ENOENT) {
      log_main(tprintf("netmare: Unable to open database file '%s': %s",
               db_file, strerror(errno)));
      exit(0);
    }
    log_main("Creating new database from scratch.");
    create_db();
    init_config();
    open_game();
  } else {
    db_read_file=f;
    log_main(tprintf("Reading file '%s'...", db_file));

    log_main(tprintf(" Date: %s", strtime(fdate(db_file))));
    log_main(tprintf(" Size: %ld bytes", fsize(db_file)));
  }
}

/* read file run/etc/portmap and open all listed ports */
void open_ports()
{
  static char *ports[]={"game", "admin", "http", "mesg", NULL};

  DESC *d;
  FILE *f;
  char buf[256], *r, *s;
  int a, type, port;

  /* read portmap file */
  if((f=fopen("etc/portmap", "r"))) {
    while(ftext(buf, 256, f)) {
      if((s=strchr(buf, '#')))
        *s='\0';
      if(!*buf)
        continue;

      /* determine type of port to open */
      s=buf;
      while(isspace(*s))
        s++;
      if(!(r=strspc(&s)))
        continue;
      for(type=0;ports[type];type++)
        if(!strcasecmp(ports[type], r))
          break;
      if(!ports[type])
        continue;

      /* find port value */
      if(!(r=strspc(&s)))
        continue;
      if(isdigit(*r)) {
        port=atoi(r);
        if(port < 1 || port > 65535) {
          log_error(tprintf("etc/portmap: Invalid %s port %d.", ports[type],
                    port));
          continue;
        }
      } else
        port=tinyport;

      /* scan current ports to see if it's already open */
      for(d=descriptor_list;d;d=d->next)
        if((d->flags & C_SOCKET) && d->port == port)
          break;

      /* open port */
      if(!d) {
        d=make_socket(port, type);

        /* save any available port options */
        for(a=0;a < ENV_SIZE && (r=strspc(&s));a++)
          WPTR(d->env[a], r);
      }
    }
    fclose(f);
  }

  /* open default port for tinymare if it isn't open already */
  for(d=descriptor_list;d;d=d->next)
    if((d->flags & C_SOCKET) && d->port == tinyport)
      break;
  if(!d && !make_socket(tinyport, SOCKET_GAME)) {
    log_error("Cannot open requested port. Game aborted.");
    exit(1);
  }

#ifndef __CYGWIN__  /* Cygwin-32 does not support lowlevel DNS queries */

  /* open connection for domain name server queries */
  for(d=descriptor_list;d;d=d->next)
    if(d->flags & C_DNS) {
      resolv=d->descriptor;
      break;
    }
  if(!d && (resolv=dns_open()) != -1) {
    d=initializesock(resolv);
    d->concid=0;
    d->flags|=C_DNS;
    d->state=-1;
    SPTR(d->addr, "localhost");
    nstat[NS_FILES]++;
    log_io(tprintf("DNS Resolver Initialized. desc->%d", resolv));
  }
#endif
}

void open_game()
{
  DESC *d;

  /* announce that the game is back up */
  if(!(boot_flags & BOOT_QUIET))
    log_main(tprintf("%s Online.", MUD_NAME));

  /* reconnect any players already on */
  for(d=descriptor_list;d;d=d->next) {
    if(d->player >= db_top || (d->player != NOTHING && Typeof(d->player) !=
       TYPE_PLAYER)) {
      output2(d, "*** Connection Reset ***\n\n");
      d->state=0;
      d->player=NOTHING;
      d->flags&=C_LINEMODE|C_REMOTE;
      d->flags|=C_RECONNECT|C_LOGOFF;
    } else if(d->player != NOTHING && (d->flags & C_WAITCONNECT)) {
      if(!(boot_flags & BOOT_QUIET))
        output2(d, "*** Reconnecting ***\n");
      d->flags&=~C_WAITCONNECT;
      d->flags|=C_CONNECTED;
    } else
      d->connected_at=now;
  }

  if(boot_flags & BOOT_QUIET)
    log_main(tprintf("%s Online.", MUD_NAME));

  /* initialize the CPU Usage counter */
  update_cpu_percent();

  /* initialize virtual timer every 10 seconds for infinite-loop detection */
  alarm(10);

  /* transfer program to main runtime routine */
  boot_flags=0;
  dberrcheck=0;
  game_online=1;
}

void close_game()
{
#ifdef __CYGWIN__
  char buf[1024];
#endif

  alarm(0);
  log_main("Server shutting down.");

  if(db_readonly) {
    /* Print a warning message and disconnect all players */
    log_main("Warning: Game running in read-only mode. "
             "Database will not be saved.");
    boot_flags |= BOOT_RESTART;
  }

  close_sockets();

  /* Save the database (only if not in read-only mode) */
  if(!db_readonly)
    dump_database();

  /* Restart Netmare if @rebooting */
  if(!(boot_flags & BOOT_RESTART) && shutdown_flag == 2) {
    log_main("Restarting TinyMare.");
#ifdef __CYGWIN__
    /* remove any double quotes surrounding pathname */
    if(*mainarg[0] == '"')
      mainarg[0][strlen(mainarg[0]++)-1]='\0';

    /* convert win32 path to posix to be used with execv */
    cygwin_conv_to_posix_path(mainarg[0], buf);

    execv(buf, mainarg);
    perror(buf);
#else
    execv(mainarg[0], mainarg);
    perror(mainarg[0]);
#endif
  }

  /* Remove temporary files */
  unlink("logs/socket_table");

  exit(shutdown_flag == 2 ? 2:0);
}

void close_sockets()
{
  FILE *f;
  DESC *d, *dnext;
  int a, b, num=0;

  /* open file to save socket table information to */
  if(shutdown_flag != 2 || (boot_flags & BOOT_RESTART) ||
     !(f=fopen("logs/socket_table", "wb")))
    f=NULL;

  /* count number of descriptors & print shutdown message */
  for(d=descriptor_list;d;d=dnext) {
    dnext=d->next;
    if(d->flags & (C_SOCKET|C_DNS)) {
      fcntl(d->descriptor, F_SETFD, 0);
      num++;
      continue;
    }

    if(d->state == 33) {  // Send reboot/shutdown signal to Remotelink server
      if(!f)
        output2(d, tprintf("$%d\n", (shutdown_flag == 2)?18:19));
    } else if(!(d->flags & C_SERVICE) && xcpu) {
      output2(d, "\a\e[1;31m*** The sky is falling!!\n");
      output2(d, "\a*** Trying Emergency Save.\e[m\n");
    } else if(!(d->flags & C_SERVICE) && !(boot_flags & BOOT_QUIET)) {
      if(f)
        output2(d, tprintf("%s Rebooting.\n", MUD_NAME));
      else
        output2(d, "Going Down - Bye.\n");
    }

    if(!f)
      shutdownsock(d);
    else {
      process_output(d);
      fcntl(d->descriptor, F_SETFD, 0);
    }

    num++;
  }

  if(!f)
    return;

  b=num;
  /* write socket table header (for reboot) */
  putnum(f, getpid());
  putlong(f, mud_up_time);
  putnum(f, boot_flags);
  putnum(f, num);
  putchr(f, ENV_SIZE);
  putchr(f, FIGHT_SUMMARY);
  putchr(f, NET_STATS);
  for(a=0;a<NET_STATS;a++)
    putnum(f, nstat[a]);

  for(;num>0;num--) {
    d=descriptor_list;
    for(a=1;a<num;a++)
      d=d->next;

    /* write connection list database */
    putnum(f, d->descriptor);
    putnum(f, d->concid);
    putnum(f, d->player);
    putnum(f, d->flags);
    putlong(f, d->login_time);
    putlong(f, d->last_time);
    putlong(f, d->connected_at);
    for(a=0;a<ENV_SIZE;a++)
      putstring(f, d->env[a]);
    putnum(f, d->state);
    putshort(f, d->socket);
    putchr(f, d->type);
    putnum(f, d->timer);
    putshort(f, d->rows);
    putshort(f, d->cols);
    putchr(f, d->term);
    putnum(f, d->cmds);
    putnum(f, d->input);
    putnum(f, d->output);
    putnum(f, d->ipaddr);
    putshort(f, d->port);
    putstring(f, d->addr);

    /* summary information */
    for(a=0;a<FIGHT_SUMMARY;a++)
      putnum(f, d->summary[a]);
  }
  putnum(f, NOTHING);

  putnum(f, b);  /* checksum */
  fclose(f);
}

int open_sockets()
{
  DESC *d, *k;
  FILE *f=NULL;
  int a, b, c, env, summary, num, desc;

  if(!(f=fopen("logs/socket_table", "rb")))
    return 0;

  /* This must be the same process id# */
  if(getpid() != getnum(f)) {
    fclose(f);
    unlink("logs/socket_table");  /* remove socket table */
    return 0;
  }

  mud_up_time=getlong(f);
  boot_flags=getnum(f);
  num=getnum(f);
  env=fgetc(f);
  summary=fgetc(f);
  a=fgetc(f);
  for(b=0;b<a;b++) {
    if(b < NET_STATS)
      nstat[b]=getnum(f);
    else
      getnum(f);
  }
  nstat[NS_EXEC]++;
  nstat[NS_DBSAVE]++;

  for(a=0;a<=num;a++) {
    if((desc=getnum(f)) == NOTHING)
      break;
    fcntl(desc, F_SETFD, 1);
    d=initializesock(desc);
    d->concid=getnum(f);
    d->player=getnum(f);
    d->flags=getnum(f);
    d->login_time=getlong(f);
    d->last_time=getlong(f);
    d->connected_at=getlong(f);
    for(c=0;c<env;c++)
      getstring(f, d->env[(c >= ENV_SIZE)?ENV_SIZE-1:c]);
    d->state=getnum(f);
    d->socket=getshort(f);
    d->type=fgetc(f);
    d->timer=getnum(f);
    d->rows=getshort(f);
    d->cols=getshort(f);
    d->term=fgetc(f);
    d->cmds=getnum(f);
    d->input=getnum(f);
    d->output=getnum(f);

    d->ipaddr=getnum(f);
    d->port=getshort(f);
    getstring(f, d->addr);

    /* load summary information */
    for(c=0;c<summary;c++) {
      if(c < FIGHT_SUMMARY)
        d->summary[c]=getnum(f);
      else
        getnum(f);
    }

    if(d->flags & C_CONNECTED) {
      d->flags&=~C_CONNECTED;
      d->flags|=C_WAITCONNECT;
    }
  }

  b=getnum(f);
  fclose(f);

  /* ensure no two processes will open this file */
  unlink("logs/socket_table");

  /* socket table checksum is different */
  if(a != b) {
    boot_flags &= ~BOOT_QUIET;
    log_main("New TinyMare version information loaded.");
    log_main("Closing all connections and restarting.");
    exit(2);
  }

  /* match file descriptors with contentrators & parents */
  for(d=descriptor_list;d;d=d->next) {
    if(!(d->flags & C_REMOTE))
      continue;
    for(k=descriptor_list;k;k=k->next)
      if(k->descriptor == d->descriptor && !(k->flags & C_REMOTE)) {
        d->parent=k;
        break;
      }
  }

  return 1;
}

/* Signal Handling */

/* protects against infinite loops in erroneous coding */
void alarm_handler(int sig)
{
  static long last_alarm;
  long tt=time(NULL);
  sigset_t set;

  /* Restart Alarm */
  signal(sig, alarm_handler);
  sigemptyset(&set);
  sigprocmask(SIG_SETMASK, &set, NULL);
  alarm(10);

  /* Check when we were last called */
  if(tt-last_alarm > 11)  /* Must be in debugging mode */
    alarm_count=-1;
  last_alarm=tt;

  if(++alarm_count > 2)
    error_handler(sig);
}

void sig_handler(int sig)
{
  signal(sig, sig_handler);
  sig_caught = sig+128; /* 8th bit: signal triggered but not processed yet */
}

/* fork() finishes from database save */
void reaper(int sig)
{
  int status;

  /* terminate children */
  wait(&status);

  if(!WIFEXITED(status) || WEXITSTATUS(status))
    badforksave=1;
  else
    save_flag=1;

  /* reset signal for next dump */
  signal(sig, reaper);
}

void error_handler(int sig)
{
  DESC *d;
  static int crashes=0;
  char buf[8192];
  int i;

  if(sig != SIGALRM) {
    if(dberrcheck++) {
      sprintf(buf, "Database consistency check failed on #%d", game_online?
              log_dbref:speaker);
      log_error(buf);
      log_error("");
      log_error("Unrecoverable error. Aborting.");

      /* process output */
      for(d=descriptor_list;d;d=d->next)
        if(!(d->flags & C_LOGOFF))
          process_output(d);

      if(!game_online)  /* failed while loading database? exit. */
        exit(0);

      signal(sig, SIG_DFL);  /* reset to default handler and trap error. */
      return;
    }

    signal(sig, SIG_DFL);
    signal(sig, error_handler);
  }

  /* display and log nature of error */
  sprintf(buf, "Game Driver: NetMARE Runtime Malfunction, %s",
          sig==SIGALRM?"Infinite loop":(char *)strsignal(sig));
  log_error(buf);
  log_error("");
  if(!Invalid(log_dbref)) {
    sprintf(buf, "FAILED COMMAND: %s(#%d) in %s(#%d)", db[log_dbref].name,
              log_dbref, db[db[log_dbref].location].name,
              db[log_dbref].location);
    log_error(buf);
  }
  if(*log_text) {
    sprintf(buf, "LOG DATA: %s", log_text);
    log_error(buf);
  }
  log_error("");

  if(++crashes > 127)
    log_error("Too many crashes. Aborting.");
  else if(sig != SIGALRM && !ERR_SIGNAL)
    log_error("Aborting TinyMare.");
  else if(sig != SIGSEGV)
    log_error("Game Driver restarted.");
  else
    log_error("Checking database integrity...");

  /* process output */
  for(d=descriptor_list;d;d=d->next)
    if(!(d->flags & C_LOGOFF))
      process_output(d);

  /* abort program if not catching error signals */
  if(sig != SIGALRM && (!ERR_SIGNAL || crashes > 127)) {
    signal(sig, SIG_DFL);
    return;
  }

  /* segmentation faults need db consistency check */
  if(sig == SIGSEGV) {
    i=db_check();
    sprintf(buf, "%d object%s, %d byte%s RAM check 100%%", db_top,
            db_top==1?"":"s", i, i==1?"":"s");
    log_error(buf);
  }

  nstat[NS_CRASHES]++;

  for(d=descriptor_list;d;d=d->next)
    if(!(d->flags & C_LOGOFF) && (d->flags & C_CONNECTED))
      output2(d, "*** Reconnecting ***\n");

  /* restart game driver */
  serverport();
}

static void check_signals()
{
  if(!(sig_caught & 128))
    return;

  sig_caught-=128;
  if(sig_caught < 1 || sig_caught >= NSIG)
    return;

  log_error(tprintf("Game Driver: Caught Signal '%s'", strsignal(sig_caught)));

  switch(sig_caught) {
  case SIGHUP:  /* perform database dump */
    do_dump(GOD);
    break;
  case SIGUSR1:  /* reboot system */
    shutdown_flag=2;
    signal(sig_caught, SIG_IGN);
    break;
  case SIGQUIT: /* normal shutdown.. */
  case SIGTERM:
    xcpu=1;
    shutdown_flag=1;
    signal(sig_caught, SIG_IGN);
  }
}

/********************/
/* Main Server Code */
/********************/

void serverport()
{
  DESC *d, *dnext;
  static struct timeval timeout;
  fd_set input_set, output_set;
  int found=0, passthru, run_queue;
  quad timer;

  /* main server sequence */
  while(1) {
    alarm_count=0;

    if(!game_online) {
      /* Load more database objects */
      if(load_database())
        open_game();  /* Load finished */
    } else {
      check_signals();

      if(save_flag) {
        if(fsize(db_file) >= 0)
          log_main(tprintf("%s Checkpoint Done. Filesize: %ld bytes", MUD_NAME,
                   fsize(db_file)));
        save_flag=0;
        epoch_fail=0;
        halt_dump=0;
      } else if(badforksave) {
        badforksave=0;
        no_dbdump(1);
        halt_dump=0;
      }

      /* begin queue time-slice of 100ms */
      timer=get_time();
      if(time_slice-timer < 1 || time_slice-timer > 100) {
        time_slice=timer+100;
        run_queue=1;
      } else
        run_queue=0;

      process_commands();

      if(shutdown_flag && !halt_dump)  /* exit main server engine */
        break;

      /* process queue events only once per 100ms */
      if(run_queue) {
        dispatch();	/* timer events */
        flush_queue();	/* process command queue */
      }
    }
    
    /* clear input/output file descriptor sets */
    FD_ZERO(&input_set);
    FD_ZERO(&output_set);

    /* set which file descriptors are ready to read input or send output */
    for(found=0,passthru=0,d=descriptor_list;d;d=d->next) {
      if(!d->io[0])
        FD_SET(d->descriptor, &input_set);
      if(d->io[2] || (d->io[1] && !(d->flags & (C_OUTPUTOFF|C_WAITCONNECT))))
        FD_SET(d->descriptor, &output_set);
      if(d->descriptor > found)
        found=d->descriptor;
      if((d->flags & C_LOGOFF) || ((d->flags & C_REMOTE) && !d->parent))
        passthru=1;
    }

    /* select() delay is remainder of 100ms time-slice not used */
    timer=time_slice-get_time();
    timeout.tv_sec = 0;
    timeout.tv_usec = (game_online && timer > 0 && timer <= 100)?timer*1000:0;
    
    found=select(found+1, &input_set, &output_set, NULL, &timeout);
    now=time(NULL);

    if(found <= 0 && !passthru)
      continue;

    /* there are descriptors ready. go into input/output processing */

    /* process all input */
    for(d=descriptor_list;d;d=dnext) {
      dnext=d->next;
      if(!(d->flags & C_REMOTE) && FD_ISSET(d->descriptor, &input_set)) {
        if(d->flags & C_SOCKET)
          new_connection(d);
#ifndef __CYGWIN__
        else if(d->flags & C_DNS)
          dns_receive();
#endif
        else if(!process_input(d))
          shutdownsock(d);
      }
    }

    /* process all output */
    for(d=descriptor_list;d;d=dnext) {
      dnext=d->next;
      if((d->flags & C_LOGOFF) || (FD_ISSET(d->descriptor, &output_set) &&
         !process_output(d)))
        shutdownsock(d);
    }

    /* finally, remove unparented remote connections */
    for(d=descriptor_list;d;d=dnext) {
      dnext=d->next;
      if((d->flags & C_LOGOFF) || ((d->flags & C_REMOTE) && !d->parent))
        shutdownsock(d);
    }
  }

  /* shutdown server */
  close_game();
}

DESC *make_socket(int port, int type)
{
  struct sockaddr_in server;
  char buf[32];
  DESC *d;
  int s, on=1;
  
  sprintf(buf, "port %d", port);

  /* create socket */
  if((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    perror(buf);
    return NULL;
  }

  /* Set option on socket to allow port to be reused. This is mandatory to be
     able to reopen the login port when rebooting or after the game crashes,
     and players are still waiting for input from a unexisting process. */

  if(setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on)) < 0) {
    perror(buf);
    close(s);
    return NULL;
  }

  /* Bind the socket to an address:port */
  server.sin_family = AF_INET;
  server.sin_addr.s_addr = INADDR_ANY;
  server.sin_port = htons(port);
  if(bind(s, (struct sockaddr *)&server, sizeof(server)) < 0) {
    perror(buf);
    close(s);
    return NULL;
  }

  /* initialize file descriptor */
  listen(s, 128);
  d=initializesock(s);
  d->concid=++nstat[NS_LOGINS];
  d->flags|=C_SOCKET;
  d->state=-1;
  d->type=type;
  nstat[NS_FILES]++;

  /* record internet address */
  d->port=port;
  SPTR(d->addr, "localhost");

  log_io(tprintf("Socket on Port %d Initialized. desc->%d", port, s));
  return d;
}

void new_connection(DESC *w)
{
  DESC *d;
  struct sockaddr_in addr;
  int newsock, addr_len=sizeof(struct sockaddr);

  /* find incoming address */
  if((newsock=accept(w->descriptor, (struct sockaddr *)&addr, &addr_len)) < 0)
    return;

  /* initialize file descriptor */
  fcntl(newsock, F_SETFL, O_NDELAY);
  d=initializesock(newsock);
  d->concid=++nstat[NS_LOGINS];
  d->socket=w->port;
  d->type=w->type;

  /* save internet address */
  d->port=ntohs(addr.sin_port);
  d->ipaddr=addr.sin_addr.s_addr;
  SPTR(d->addr, inet_ntoa(addr.sin_addr));

  /* check if we exceed site limit */
  nstat[NS_FILES]++;
  if((unsigned int)open_files < 0x7fffffff && nstat[NS_FILES] >
     (int)(open_files-6)) {
    log_io(tprintf("User Limit Exceeded %d: %s. desc->%d", d->concid,
           d->addr, d->descriptor));
    output2(d, tprintf("Sorry, this port is at its maximum of %d users.\n\n",
            (int)(open_files-6)));
    display_textfile(d, "msgs/toomany.txt");
    shutdownsock(d);
    return;
  }

  /* check IP address for sitelock */
  if(forbidden_site(d->addr)) {
    log_io(tprintf("Refused %d: %s. desc->%d", d->concid, d->addr, newsock));
    display_textfile(d, "msgs/lockout.txt");
    shutdownsock(d);
    return;
  }

  /* Initialize for a special socket connection */
  switch(w->type) {
    case SOCKET_HTTP: d->state=37; d->flags|=C_SUPPRESS; break;
    case SOCKET_MESG: d->state=38; d->flags|=C_SUPPRESS; break;
  }

  /* query dns server for hostname information */
  if(!init_connect(d))
    d->flags|=C_RESOLVER;
  else if(d->flags & C_LOGOFF) {
    shutdownsock(d);
    return;
  }

  /* enable transmit Telnet EOR, NAWS, and TTYPE */
  if(w->type != SOCKET_HTTP && w->type != SOCKET_MESG) {
    output2(d, "\377\373\031");  // Will EOR
    output2(d, "\377\375\037");  // Do NAWS
    output2(d, "\377\375\030");  // Do TTYPE
    output2(d, "\377\372\030\001\377\360");  // Negotiate TTYPE
  }

  /* display login information to user */
  if(w->type == SOCKET_GAME) {
    output2(d, "\ec\e(K\e[1;5]\e[2J\e[H\n");
    display_textfile(d, "msgs/title.txt");
  } else if(w->type == SOCKET_ADMIN) {
    output2(d, tprintf("%s: %s (fd%d) #%d\n", mud_version,
            system_info, d->descriptor, d->concid));
    output2(d, tprintf("%s -- %s %d\n\n", mud_copyright, MUD_NAME, tinyport));
  }
}

/* initialize the new connection */
DESC *initializesock(int s)
{
  DESC *d;
  int a;

  /* allocate a new descriptor structure and zero its contents */
  d=calloc(1, sizeof(DESC));

  /* initialize data on descriptor pointer */
  d->descriptor=s;
  d->player=NOTHING;

  /* concentrator info */
  d->flags=C_LOGIN;
  d->login_time=now;
  d->last_time=now;
  d->connected_at=now;

  /* connection environment */
  for(a=0;a<ENV_SIZE;a++)
    SPTR(d->env[a], "");

  d->next=descriptor_list;
  descriptor_list=d;
  return d;
}

/* look up user hostname and identity */
int init_connect(DESC *d)
{
  char buff[256], *s;
  
  /* attempt to find hostname */
  if(!(s=dns_lookup(d->ipaddr)))
    return 0;
  
  /* get accounting information */
  get_ident(buff, d->descriptor);
  if(*buff)
    strcat(buff, "@");
  strcat(buff, s);
  SPTR(d->addr, buff);

  if(forbidden_site(d->addr)) {
    log_io(tprintf("Refused %d: %s. desc->%d", d->concid, d->addr,
           d->descriptor));
    display_textfile(d, "msgs/lockout.txt");
    d->flags|=C_LOGOFF;
    return 1;
  }

  if(!(d->flags & C_SUPPRESS))
    log_io(tprintf("Login %d: %s. desc->%d", d->concid,d->addr,d->descriptor));
  return 1;
}

int forbidden_site(char *addr)
{
  FILE *f;
  char buf[256], *r, *s, *host;

  if((f=fopen("etc/lockout", "r")) == NULL)
    return 0;
  if((host=strchr(addr, '@')))
    host++;

  while(ftext(buf, 256, f)) {
    if((s=strchr(buf, '#')))
      *s='\0';
    if(!*buf)
      continue;

    for(s=buf;*s;s=r) {
      while(isspace(*s))
        s++;
      if(!*s)
        break;
      r=s;
      while(*r && !(isspace(*r) || *r == ','))
        r++;
      *r++='\0';

      /* check against: 1) full address, and 2) hostname only */
      if(wild_match(addr, s, 0, 0)) {
        fclose(f);
        return 1;
      } if(host && wild_match(host, s, 0, 0)) {
        fclose(f);
        return 1;
      }
    }
  }

  fclose(f);
  return 0;
}


/* reads input from connected players */
int process_input(DESC *d)
{
  struct io_queue *new;
  unsigned char buf[4096], temp[8192];
  int a, b, len;

  /* read data & check that socket is still alive */
  if((len = read(d->descriptor, buf, 4096)) <= 0) {
    if(errno == EAGAIN)
      return -1;
    return 0;
  }

  nstat[NS_INPUT]+=len;
  d->input+=len;

  if(!d->raw_input) {
    *temp=0;
    b=0;
  } else {
    strcpy(temp, d->raw_input);
    b=strlen(temp);
  }

  /* filter input */
  for(a=0;a < len;a++) {
    /* Process telnet command characters */
    if(buf[a] == IAC && a < 4094) {
      a+=telnet_iac(d, buf+a+1, len-a-1);
      continue;
    }

    if(buf[a] == 8 || buf[a] == 127) {  /* backspace */
      if(b > 0)
        b--;
      continue;
    }
    if(buf[a] == 10) {  /* newline */
      /* remove following spaces */
      while(b > 0 && temp[b-1] == ' ')
        b--;
      temp[b]=0;

      /* save command to input queue */
      new=malloc(sizeof(struct io_queue)+b+1);
      strcpy(new->text, temp);
      if(d->io_last[0])
        d->io_last[0]->next=new;
      else
        d->io[0]=new;
      d->io_last[0]=new;
      new->next=NULL;

      *temp=0;
      b=0;
      continue;
    }
    if(buf[a] < 32 || buf[a] > 127 || b > 7999)
      continue;
    temp[b++]=buf[a];
  }
  temp[b]=0;

  /* save input for next read */
  WPTR(d->raw_input, temp);

  return 1;
}

/* writes output to connected players */
int process_output(DESC *d)
{
  char buffer[32768], *s=buffer;
  struct io_queue *ptr, *pnext, *new;
  int a, len, amt, size=0;
  
  if((d->flags & C_REMOTE) && !d->parent)
    return 0;

  /* process server output then regular game output */
  for(a=2;a>=1;a--) {
    if(a == 1 && (d->flags & (C_OUTPUTOFF|C_WAITCONNECT)))
      break;
    ptr=d->io[a];
    while(ptr || size) {
      if(ptr) {
        pnext=ptr->next;
        amt=strlen(ptr->text);
      } else {
        amt=0;
        pnext=NULL;
      }

      if(size+amt > 32767 || !ptr) {
        if(a == 1)
          d->output_size-=size;
        if((len=write(d->descriptor, buffer, size)) < size) {
          if(len != -1 || errno == EAGAIN) {
            /* move unsent data to highest priority on queue */
            if(len < 0)
              len=0;
            new=malloc(sizeof(struct io_queue)+strlen(buffer+len)+1);
            strcpy(new->text, buffer+len);
            new->next=d->io[2];
            d->io[2]=new;
            if(!d->io_last[2])
              d->io_last[2]=new;
            return -1;
          }
          return 0;
        }
        nstat[NS_OUTPUT]+=len;
        d->output+=len;
        size=0;
        s=buffer;
      }

      if(!ptr)
        break;
      strcpy(s, ptr->text);
      size+=amt;
      s+=amt;

      free(ptr);
      if(!(d->io[a]=ptr=pnext))
        d->io_last[a]=NULL;
    }
  }

  return 1;
}

/* Telnet "Is A Command" Character Processing */
int telnet_iac(DESC *d, unsigned char *s, int len)
{
  unsigned char buf[32];
  int i, j;

  if(*s < 236)
    return 0;

  /* Telnet EOR-Style Prompts */
  if(*s == DO && *(s+1) == TELOPT_EOR)
    d->flags|=C_LINEMODE;

  /* Handle Subnegotiation */
  else if(*s == SB) {
    for(i=1,j=0;i < len && j < 32;i++) {
      if(*(s+i) == IAC && *(s+i+1) == SE) {
        i++;
        break;
      } if(*(s+i) == IAC && *(s+i+1) == IAC) {
        buf[j++]=IAC;
        i++;
      } else
        buf[j++]=*(s+i);
    }

    /* Select Terminal Type */
    if(buf[0] == TELOPT_TTYPE && !buf[1]) {
      /* Convert to lowercase */
      buf[j]='\0';
      for(j--;j>1;j--)
        buf[j]=tolower(buf[j]);

      /* Match type. Supports maximum features at type 1, minimum at 4. */
      if(!strcmp("mare", buf+2))
        d->term=1;
      else if(!strcmp("linux", buf+2))
        d->term=2;
      else if(match_word("xterm ansi vt100 vt102 vt200 vt220 vt320 screen "
              "crt zmud", buf+2))
        d->term=3;
      else if(!strcmp("dumb", buf+2))
        d->term=4;
      else if(buf[2])
        log_io(tprintf("Session %d: Unknown terminal type '%s'.",
               d->concid, buf+2));
    }

    /* Now Adjust Window Size */
    else if(buf[0] == TELOPT_NAWS && j > 4) {
      if((d->cols=(buf[1]<<8)|buf[2]) > 500)
        d->cols=500;
      if((d->rows=(buf[3]<<8)|buf[4]) > 500)
        d->rows=500;
    }

    return i;
  }

  /* Are you there? */
  else if(*s == AYT)
    output2(d, tprintf("\e[1m[%s %d]\e[m\n", MUD_NAME, d->socket));

  return (*s < 251)?1:2;
}

/* close remote-user socket connection */
void shutdownsock(DESC *d)
{
  DESC *ptr;
  dbref player=d->player;
  int flags=d->flags;

  /* prevent people from disconnecting while database is still loading */
  if(!game_online)
    return;

  /* if not reconnecting, clear everything */
  if(!(d->flags & C_RECONNECT)) {
    if(!Invalid(player)) {
      if(terminfo(player, TERM_SOUND))
        output2(d, "\e[0;32;0;7;36t");
      d->flags&=~C_OUTPUTOFF;
      if(terminfo(player, TERM_ANSI))  /* reset screen */
        output2(d, "\e7\e[r\e8\e[A\n");
      display_textfile(d, "msgs/leave.txt");
      process_output(d);
      log_io(tprintf("Disconnect %d: %s(#%d). desc->%d", d->concid,
             db[player].name, player, d->descriptor));
      if(d->flags & C_CONNECTED)
        announce_disconnect(player);
    } else {
      process_output(d);
      if(!(d->flags & C_SUPPRESS))
        log_io(tprintf("%sLogout %d: %s. desc->%d", d->flags&C_PARENT?
               "Concentrator ":"", d->concid, d->addr, d->descriptor));
    }
  } else {
    int a;

    d->flags&=~(C_OUTPUTOFF|C_CONNECTED);
    process_output(d);
    for(a=0;a<ENV_SIZE;a++)  /* clear environment */
      WPTR(d->env[a], "");

    if(!Invalid(player)) {
      log_io(tprintf("Reconnect %d: %s(#%d). desc->%d", d->concid,
             db[player].name, player, d->descriptor));
      announce_disconnect(player);
    } else
      log_io(tprintf("Relog %d: %s. desc->%d", d->concid,
             d->addr, d->descriptor));
  }

  /* clear children */
  if(flags & C_PARENT)
    for(ptr=descriptor_list;ptr;ptr=ptr->next)
      if(ptr->parent == d)
        ptr->parent=NULL;

  /* reset resolver file-id */
  if(flags & C_DNS)
    resolv=-1;

  d->player=NOTHING;

  /* Erase half-created characters, @nuked players, or Guests */
  if(!Invalid(player) && Typeof(player) == TYPE_PLAYER &&
     (Guest(player) || (flags & C_DESTROY) || !db[player].data->last)) {
    if(!(flags & C_CONNECTED))
      log_io(tprintf("Autonuke: %s(#%d).", db[player].name, player));
    destroy_player(player);
  }

  /* logging off */
  if(!(d->flags & C_RECONNECT)) {
    if(d->flags & C_REMOTE) {
      for(ptr=descriptor_list;ptr;ptr=ptr->next)
        if(d->parent == ptr) {
          output2(ptr, tprintf("-%d\n", d->concid));
          break;
        }
    } else {
      shutdown(d->descriptor, 2);
      close(d->descriptor);
      nstat[NS_FILES]--;
    }
    for(ptr=descriptor_list;ptr;ptr=ptr->next)
      if(ptr->next == d)
        break;
    if(ptr)
      ptr->next=d->next;
    else
      descriptor_list=d->next;

    free_io(d);
    free(d);
  } else {
    /* lets reconnect */
    d->state=0;
    d->connected_at=now;
    d->flags&=C_LINEMODE|C_REMOTE;
    d->flags|=C_LOGIN;
  }
}

void free_io(DESC *d)
{
  struct io_queue *ptr, *next;
  int a;

  for(a=0;a<3;a++)
    for(ptr=d->io[a];ptr;ptr=next) {
      next=ptr->next;
      free(ptr);
    }

  for(a=0;a<ENV_SIZE;a++)
    free(d->env[a]);
  free(d->raw_input);
}

/* adds <msg> to the output queue for a particular descriptor */
void output(DESC *d, char *msg)
{
  struct io_queue *new, *ptr, *pnext;
  char *s=term_emulation(d, msg);
  int len=strlen(s);

  new=malloc(sizeof(struct io_queue)+len+1);
  strcpy(new->text, s);
  if(d->io_last[1])
    d->io_last[1]->next=new;
  else
    d->io[1]=new;
  d->io_last[1]=new;
  new->next=NULL;
  d->output_size+=len;
  if(d->output_size <= MAX_OUTPUT)
    return;

  /* flush beginning of output */

  for(ptr=d->io[1]; d->output_size > MAX_OUTPUT && ptr; ptr=pnext) {
    pnext=ptr->next;
    d->output_size-=strlen(ptr->text);
    free(ptr);
  }
  if(!(d->io[1]=ptr))
    d->io_last[1]=NULL;

  /* add message to start of queue */
  s=term_emulation(d, "\256Output Flushed\257\n");
  new=malloc(sizeof(struct io_queue)+18);
  strcpy(new->text, s);
  new->next=d->io[1];
  d->io[1]=new;
  if(!d->io_last[1])
    d->io_last[1]=new;
  d->output_size+=17;  /* strlen(s) */
}

/* adds <msg> to server output queue */
void output2(DESC *d, char *msg)
{
  struct io_queue *new;
  char *s=term_emulation(d, msg);
 
  new=malloc(sizeof(struct io_queue)+strlen(s)+1);
  strcpy(new->text, s);
  if(d->io_last[2])
    d->io_last[2]->next=new;
  else
    d->io[2]=new;
  d->io_last[2]=new;
  new->next=NULL;
}

void raw_notify(dbref player, char *msg)
{
  DESC *d=match_desc(player);
  char buf[8192], *s;

  if(!d || !(db[player].flags & PLAYER_CONNECT))
    return;

  if(!(s=memccpy(buf, msg, 0, 8000)))
    s=buf+8000;
  else
    s--;
  *s++='\n';
  *s='\0';
  output(d, buf);
}

void process_commands()
{
  DESC *d, *dnext;
  struct io_queue *ptr;
  int count;

  /* go through the entire list of users online */
  for(d=descriptor_list;d;d=dnext) {
    dnext=d->next;
    if(!d->io[0] || (d->flags & (C_LOGOFF|C_INPUTOFF|C_RESOLVER)) ||
       ((d->flags & C_REMOTE) && !d->parent))
      continue;

    /* determine max commands per user */
    if(d->flags & C_PARENT)
      count=250;
    else if((d->flags & C_CONNECTED) && power(d->player, POW_SECURITY))
      count=50;  /* wizards can execute up to 50 commands per timeslice */
    else
      count=1;

    /* process any pending commands */
    for(;count;count--) {
      if(!d->io[0] || (d->flags & C_INPUTOFF))
        break;
      ptr=d->io[0];
      if(!(d->io[0]=ptr->next))
        d->io_last[0]=NULL;

      if(!do_command(d, ptr->text)) {
        shutdownsock(d);
        break;
      }

      free(ptr);
    }
  }
}

int do_command(DESC *d, char *command)
{
  char *s;
  int a;

  /* reset environment %0 through %9 */
  for(a=0;a<10;a++)
    *wptr[a]='\0';

  /* remove leading spaces */
  if(d->state != 32 && d->state != 36)
    while(*command && *command == ' ')
      command++;

  /* command unidles player from over 1 hour */
  if(now-d->last_time >= 3600L && (d->flags & C_CONNECTED)) {
    s=time_format(TMS, now-d->last_time);
    speaker=d->player;
    log_io(tprintf("Game: %s has unidled. (%s)", db[d->player].name, s));
    do_cnotify(d->player, tprintf("\e[1mNotify:\e[m %s has unidled. "
               "\e[1;36m(\e[37m%s\e[36m)\e[m", db[d->player].name, s), 0);
    if(*(s=atr_get(d->player, A_AUNIDLE))) {
      sprintf(wptr[0], "%ld", now-d->last_time);
      parse_que(d->player, s, d->player);
      *wptr[0]='\0';
    }
  }

  d->last_time=now;
  d->cmds++;

  log_dbref=NOTHING;
  sprintf(log_text, "Id#%d: %s", d->concid, command);

  /* priority commands */
  if(d->state < 31 || d->state > 32) {
    if(!strcmp(command, "QUIT"))
      return 0;
    if(d->state > 4 && !strcmp(command, "RECONNECT")) {
      process_output(d);
      output2(d, "*** Reconnecting ***\n\n");
      d->flags|=C_RECONNECT;
      return 0;
    }
  }

  /* print output prefix */
  if(d->state == 10 && *(s=atr_parse(d->player, d->player, A_PREFIX))) {
    strcat(s, "\n");
    output(d, s);
  }

  /* process individual command */
  if(!server_command[d->state].fun(d, command))
    return 0;

  if(d->state == 10) {
    /* print output suffix */
    if(*(s=atr_parse(d->player, d->player, A_SUFFIX))) {
      strcat(s, "\n");
      output(d, s);
    }

    /* display prompt */
    if(*(s=atr_parse(d->player, d->player, A_PROMPT))) {
      if(!strcasecmp(s, "on"))
        s=">";
      s=tprintf("%s %s", s, PROMPT(d));
      if(d->flags & C_OUTPUTOFF)
        output2(d, s);
      else
        output(d, s);
    }
  }

  return 1;
}

DESC *match_desc(dbref player)
{
  DESC *d;

  for(d=descriptor_list;d && d->player != player;d=d->next);
  return d;
}

// Flush descriptor I/O buffers.
// Index=-1: All Buffers
// Index= 0: Input
// Index= 1: Output
// Index= 2: Server Output
//
void flush_io(DESC *d, int index)
{
  struct io_queue *ptr, *next;
  int a;

  for(a=0;a<3;a++) {
    if(index != a && index != -1)
      continue;

    /* clear text buffer */
    for(ptr=d->io[a];ptr;ptr=next) {
      next=ptr->next;
      free(ptr);
    }
    d->io_last[a]=d->io[a]=NULL;
  }

  /* Flush unread input (necessary during C_INPUTOFF) */
  if(index < 1) {
    struct timeval timeout={0, 0};
    fd_set input_set;
    char buf[4096];

    FD_ZERO(&input_set);
    FD_SET(d->descriptor, &input_set);
    while(select(d->descriptor+1, &input_set, (fd_set *)NULL, (fd_set *)NULL,
          &timeout) > 0)
      read(d->descriptor, buf, 4096);
  }
}

/* clears input/output buffers on individual player */
void do_flush(dbref player, char *arg1, char *arg2)
{
  const char *type[3]={"Input", "Output", "Server"};
  dbref a, thing=match_thing(player, arg1, MATCH_EVERYTHING);
  char *r, *s=arg2;
  DESC *d;

  /* make sure we have a valid player */
  if(thing == NOTHING)
    return;
  if(!(d=match_desc(thing))) {
    notify(player, "That player is not logged on.");
    return;
  }

  /* if no options present, flush all */
  if(!*arg2) {
    flush_io(d, -1);
    if(!Quiet(player))
      notify(player, "Buffers flushed.");
    return;
  }

  while((r=strspc(&s))) {
    for(a=0;a<3;a++)
      if(string_prefix(type[a], r))
        break;
    if(a == 3)
      notify(player, tprintf("No such option '%s'.", r));
    else {
      flush_io(d, a);
      if(!Quiet(player))
        notify(player, tprintf("%s%s flushed.", type[a], a==2?" output":""));
    }
  }
}

/* open a connection to another port from within mare */
DESC *open_outbound(char *host, int port)
{
  DESC *d;
  struct sockaddr_in addr;
  struct hostent *hent;
  int sock;

  if(port < 1 || port > 65535)
    return 0;
  if((unsigned int)open_files < 0x7fffffff &&
     nstat[NS_FILES] >= (int)(open_files-6))
    return 0;

  if((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    perror("open_outbound");
    return 0;
  }

  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
  if(!(hent=gethostbyname(host)))  /* blocks server if cannot resolve */
    addr.sin_addr.s_addr = inet_addr(host);
  else
    addr.sin_addr.s_addr = *(long *)hent->h_addr_list[0];

  if(connect(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
    close(sock);
    return 0;
  }

  nstat[NS_FILES]++;
  d=initializesock(sock);
  d->concid=++nstat[NS_LOGINS];
  d->flags|=C_OUTBOUND;
  d->state=-1;

  /* save internet address */
  d->port=port;
  SPTR(d->addr, host);

  log_io(tprintf("Outbound %d: %s port %d. desc->%d", d->concid, host,
         port, d->descriptor));
  return d;
}
