/* io/user.c */
/* Commands dealing with user lists and connections */

#include "externs.h"

/* whoflag format identifiers */

#define WF_RJUST  0x1	// Right-justify column data
#define WF_WIZARD 0x2	// POW_WHO required to view contents
#define WF_EXTEND 0x4	// Column data can extend width by 2 characters
#define WF_COMBAT 0x8	// Column available only if combat system is on
#define WF_GUILD  0x10	// Reserve 1 space for a prefix (*)

void do_who(dbref player, char *flags, char *names)
{
  static struct who_list {
    char *name;
    char len, shortlen;
    char format;
  } who_flags[]= {
     {"Name",       16, 9,  0},
     {"Alias",      10, 5,  0},
     {"Flags",      4,  4,  0},
     {"Online",     9,  9,  WF_RJUST},
     {"Idle",       4,  4,  WF_RJUST},
     {"Concid",     6,  4,  WF_WIZARD|WF_EXTEND},
     {"Port",       4,  4,  WF_WIZARD|WF_EXTEND},
     {"Desc",       4,  4,  WF_WIZARD|WF_EXTEND},
     {"Cmds",       5,  5,  WF_WIZARD|WF_EXTEND},
     {"Input",      6,  6,  WF_WIZARD|WF_EXTEND},
     {"Output",     7,  7,  WF_WIZARD|WF_EXTEND},
     {"Level",      10, 5,  WF_COMBAT},
     {"Experience", 18, 10, WF_COMBAT},
     {"Class",      10, 6,  0},
     {"Race",       11, 6,  WF_COMBAT},
     {"Gender",     6,  1,  0},
     {"Chapter",    8,  2,  WF_COMBAT},
     {"Occupation", 11, 5,  WF_GUILD|WF_COMBAT},
     {"Doing",      36, 22, 0},
     {"Where",      36, 22, 0},
     {"Address",    15, 15, WF_WIZARD},
     {"Hostname",   29, 20, WF_WIZARD},
     {"Poll",       0,  0,  0},
     {0}};

  /* Default @whoflag lists. Game adds to the @whoflag list for each new column
     that fits within the player's screen size. Numbers appearing a second time
     switch the flag to the long format. A third 0 terminates the list. */

  static struct {
    int cols;		// User must have at least this many columns on screen.
    int users;		// Number of players online.
    int wizmode;	// 1=Matches only if player has POW_WHO.
    int combat_long, combat_short;	// Bits shown if combat is turned on.
    int normal_long, normal_short;	// If combat is turned off.
  } def_bitmask[]={
    {200,   0, 1,  0x020800, 0x64871d,  0x242021, 0x4887de},
    {192,   0, 1,  0x020800, 0x64871d,  0x242001, 0x4887de},
    {180,   0, 1,  0x020800, 0x64871d,  0x242001, 0x48871e},
    {160,   0, 1,  0x020800, 0x64871d,  0x242001, 0x40871e},
    {132,   0, 1,  0x020800, 0x64871d,  0x040001, 0x60071e},
    {  0,   0, 1,  0x020800, 0x448019,  0x040001, 0x40001a},   // Wiz Default
    {  0, 175, 0,  0x000000, 0x400801,  0x000001, 0x400000},
    {  0,   0, 0,  0x020800, 0x448019,  0x040001, 0x40001a}};  // Default

  DESC *d;
  int count=0, header=1, cols=get_cols(NULL, player);
  int line=0, total=0, hidden=0, admin=0;
  int a, b, pow, small, len=0, groups, lower=0, upper=0;
  char *r, *s, buf[1024], temp[32], format[32];
  dbref obj, *namelist=NULL;

  /* request for flag list */
  if(*flags == '?') {
    *buf=0;
    for(a=0;who_flags[a].name;a++) {
      if(who_flags[a].format & WF_COMBAT)
        continue;
      if(*buf)
        strcat(buf, ", ");
      strcat(buf, who_flags[a].name);
    }
    notify(player, tprintf("List of possible flags: %s", buf));
    return;
  }

  /* determine layout flags */
  strmaxcpy(buf, *flags ? flags:atr_get(player, A_WHOFLAGS), 1024);
  if(!*buf)
    strmaxcpy(buf, atr_get(0, A_WHOFLAGS), 1024);
  s=buf;
  while((r=parse_up(&s, ' '))) {
    small=0;
    for(a=0;who_flags[a].name;a++)
      if(!(who_flags[a].format & WF_COMBAT) &&
         string_prefix(who_flags[a].name, r)) {
        if(*r >= 'a' && *r <= 'z')
          lower|=1<<a;
        else
          upper|=1<<a;
        small=1;
      }
    if(!small) {
      notify(player, tprintf("Unknown who flag '%s'.", r));
      notify(player, "Type 'who ?' for a list of available flags.");
      return;
    }
  }

  /* if no flags specified, use default bitmask */
  if(!upper && !lower) {
    /* quick count of users online */
    for(a=0,d=descriptor_list;d;d=d->next)
      if(d->flags & C_CONNECTED)
        a++;

    /* scan for default bitmask */
    pow=power(player, POW_WHO);
    for(b=0;def_bitmask[b].cols;b++)
      if(cols >= def_bitmask[b].cols && a >= def_bitmask[b].users && 
         (!def_bitmask[b].wizmode || pow))
        break;

    /* set defaults for long & short forms */
    upper=def_bitmask[b].normal_long;
    lower=def_bitmask[b].normal_short;
  }

  /* determine # groups in output */
  if(cols < 4)
    cols=4;
  for(a=0;who_flags[a].name;a++) {
    if(!who_flags[a].len)
      continue;
    if(upper & (1<<a))
      len+=who_flags[a].len+2;
    else if(lower & (1<<a))
      len+=who_flags[a].shortlen+2;
  }
  len+=2;
  groups=(cols+2)/len;
  if(groups < 1)
    groups=1;

  /* process specified names */
  if(*names) {
    s=names;
    while((r=parse_up(&s, ' '))) {
      if((obj=lookup_player(r)) <= NOTHING)
        notify(player, tprintf("No such player '%s'.", r));
      else if(!inlist(namelist, obj))
        PUSH_L(namelist, obj);
    }
    if(!namelist)
      return;
  }

  /* display the poll */
  default_color=9;
  if(((upper|lower) & 0x400000) && *(s=parse_function(player, player,
     atr_get(0, A_DOING))))
    notify(player, tprintf("\e[1;31mQuote of the Day: %s\e[m", s));
  default_color=7;
  if((upper|lower) == 0x400000) {
    if(!*s)
      notify(player, "There has been no Quote of the Day set.");
    CLEAR(namelist);
    return;
  }

  *buf=0;
  d=descriptor_list;

  /* main loop */
  while(d) {
    /* Make sure this player has connected to the game. */
    if(!(d->flags & C_CONNECTED)) {
      d=d->next;
      continue;
    }
    
    /* check namelist for player */
    if(namelist && !inlist(namelist, d->player)) {
      d=d->next;
      continue;
    }

    if(!header) {
      total++;
      if(!could_doit(player, d->player, A_HLOCK)) {
        hidden++;
        if(!controls(player, d->player, POW_WHO)) {
          d=d->next;
          continue;
        }
      }
      if(Immortal(d->player))
        admin++;
    }

    for(a=0;who_flags[a].name;a++) {
      /* determine column format */
      if(!who_flags[a].len)
        continue;
      if(!(small=(lower & (1<<a))) && !(upper & (1<<a)))
        continue;
      b=small?who_flags[a].shortlen:who_flags[a].len;
      if(header && !HIDDEN_ADMIN && (who_flags[a].format & WF_GUILD))
        sprintf(format, " %%-%d.%ds ", b, b);
      else if(!header && (who_flags[a].format & WF_EXTEND))
        sprintf(format, "%%-%d.%ds", b+2, b+2);
      else
        sprintf(format, (who_flags[a].format & WF_RJUST)?"%%%d.%ds  ":
                "%%-%d.%ds  ", b, b);

      if(header) {
        strcat(buf, tprintf(format, ((who_flags[a].format & WF_GUILD) &&
               small)?"Guild":who_flags[a].name));
        continue;
      } if((who_flags[a].format & WF_WIZARD) && !controls(player,
           d->player, POW_WHO)) {
        strcat(buf, tprintf(format, "?"));
        continue;
      }

      switch(a) {
      case 0: s=db[d->player].name; break;
      case 1: s=atr_get(d->player, A_ALIAS); break;
      case 2: 
        temp[0]=!could_doit(player, d->player, A_HLOCK)?'H':' ';
        temp[1]=(now-d->last_time >= 300 && *(atr_get(d->player, A_IDLE)))?
                'I':' ';
        temp[2]=!could_doit(player, d->player, A_PLOCK)?'P':' ';
        temp[3]=Builder(d->player)?'B':' ';
        temp[4]='\0';
        s=temp;
        break;
      case 3: s=time_format(TML, now-d->login_time); break;
      case 4: s=time_format(TMS, now-d->last_time); break;
      case 5: s=tprintf("%d", d->concid); break;
      case 6: s=tprintf("%d", d->port); break;
      case 7: s=tprintf("%d", d->descriptor); break;
      case 8: s=tprintf("%d", d->cmds); break;
      case 9: s=tprintf("%d", d->input); break;
      case 10: s=tprintf("%d", d->output); break;
      case 13: s=classnames[class(d->player)]; break;
      case 15: s=db[d->player].data->gender?"Female":"Male"; break;
      case 18:
        s=ansi_strip(atr_parse(d->player, player, A_DOING));
        for(r=s;*r;r++)  /* remove unprintable characters */
          if(*r < 32)
            *r=' ';
        break;

      case 19: s=(!controls(player, d->player, POW_EXAMINE) &&
                 (db[d->player].flags & DARK))?"You can't tell.":
                 db[db[d->player].location].name; break;
      case 20: s=ipaddr(d->ipaddr); break;
      case 21: s=d->addr; break;
      default: s="";
      }

      sprintf(buf+strlen(buf), format, s);
    }

    /* space between groups */
    if(++count < groups) {
      strcat(buf, "  ");
      if(!header)
        d=d->next;
      continue;
    }

    /* trim excess spaces */
    r=strchr(buf,'\0');
    while(*--r == ' ')
      *r='\0';

    /* print output line */
    if(header) {
      notify(player, tprintf("\e[1;4;35m%s\e[m", buf));
      header=0;
    } else {
      notify(player, tprintf("\e[%s36m%s\e[m", (line^=1)?"1;":"", buf));
      d=d->next;
    }

    count=0;
    *buf=0;
  }

  CLEAR(namelist);

  /* print last line (if incomplete) */
  if(*buf) {
    r=strchr(buf,'\0');
    while(*--r == ' ')
      *r='\0';
    notify(player, tprintf("\e[%s36m%s\e[m", (line^=1)?"1;":"", buf));
  }

  notify(player, "\e[1m\304\304\304\e[m");

  if(!power(player, POW_WHO) && RESTRICT_HIDDEN)
    total-=hidden;
  sprintf(buf, "Total Players: %d", total);
  if(admin && (!HIDDEN_ADMIN || Immortal(player)))
    strcat(buf, tprintf("   Administrators: %d", admin));
  if(hidden && (!RESTRICT_HIDDEN || power(player, POW_WHO)))
    strcat(buf, tprintf("   Hidden: %d", hidden));
  notify(player, tprintf("\e[1;32m%s.  \e[33m%s\e[m", buf, grab_time(player)));
}


void announce_disconnect(dbref player)
{
  speaker=player;

  delete_message(player, -1);
  if(db[player].data)
    db[player].data->lastoff=now;
  db[player].flags&=~PLAYER_CONNECT;

  do_cnotify(player, tprintf("\e[1mDisconnect:\e[m %s", db[player].name), 1);
  if(!Invalid(db[player].location)) {
    char *s=atr_parse(player, player, A_ODISC);
    notify_except(db[player].location, player, tprintf("%s %s",
                  db[player].name, *s?s:"has disconnected."));
  }

  global_trigger(player, A_ADISC);

  if(Guest(player))
    destroy_player(player);
}

void global_trigger(dbref who, unsigned char acmd)
{
  dbref thing, *ptr;
  char *s;

  /* players being nuked get this */
  if(Typeof(who) == TYPE_GARBAGE)
    return;

  /* universal zone commands */
  for(ptr=uzo;ptr && *ptr != NOTHING;ptr++)
    if(*(s=atr_get(*ptr, acmd)))
      parse_que(*ptr, s, who);

  /* player commands */
  if(*(s=atr_get(who, acmd)))
    parse_que(who, s, who);
  if(*(s=atr_get(db[who].location, acmd)))
    parse_que(db[who].location, s, who);

  /* room zone commands */
  for(ptr=db[mainroom(who)].zone;ptr && *ptr != NOTHING;ptr++) {
    if(db[*ptr].flags & ZONE_UNIVERSAL)
      continue;
    if(*(s=atr_get(*ptr, acmd)))
      parse_que(*ptr, s, who);
  }

  /* player contents */
  DOLIST(thing, db[who].contents)
    if(Typeof(thing) != TYPE_PLAYER && *(s=atr_get(thing, acmd)))
      parse_que(thing, s, who);

  /* room contents */
  DOLIST(thing, db[db[who].location].contents)
    if(Typeof(thing) != TYPE_PLAYER && *(s=atr_get(thing, acmd)))
      parse_que(thing, s, who);
}

void broadcast(char *msg)
{
  DESC *d;

  for(d=descriptor_list;d;d=d->next)
    if(d->flags & C_CONNECTED)
      output(d, msg);
}

void check_timeout(DESC *d)
{
  /* immortals are unsusceptable to timeout activity */
  if(!Invalid(d->player) && Immortal(d->player))
    return;

  /* first check limits on services */
  switch(d->state) {
  case 37:  // Web Server Port
    if((now - d->last_time) > 300L)
      d->flags |= C_LOGOFF;
    return;
  }

  /* then limits on valid player sessions */
  if(d->state > 2) {
    if(TIMEOUT_SESSION > 0 && (now - d->last_time) > TIMEOUT_SESSION) {
      output2(d, "*** Session Inactivity Timeout ***\n");
      log_io(tprintf("Session Inactivity %d", d->concid));
      d->flags |= C_LOGOFF;
    } else if(LIMIT_SESSION > 0 && (now - d->connected_at) > LIMIT_SESSION) {
      output2(d, "*** Time Limit Exceeded ***\n");
      log_io(tprintf("Time Limit Exceeded %d", d->concid));
      d->flags |= C_LOGOFF;
    }
    return;
  }
  
  /* now check limits at Login: prompt */

  if(TIMEOUT_LOGIN > 0 && (now - d->last_time) > TIMEOUT_LOGIN) {
    output2(d, tprintf("\nLogin expired after %d seconds.\n",
            TIMEOUT_LOGIN));
    d->flags |= C_LOGOFF;
  } else if(LIMIT_LOGIN > 0 && (now - d->connected_at) > LIMIT_LOGIN) {
    output2(d, tprintf("\nLogin time limit exceeded after %d seconds.\n",
            LIMIT_LOGIN));
    d->flags |= C_LOGOFF;
  }
}

int getidletime(dbref player)
{
  DESC *d=match_desc(player);

  if(!d)
    return -1;
  return (now - d->last_time);
}

int isbusy(dbref player)
{
  DESC *d;

  for(d=descriptor_list;d;d=d->next)
    if(d->player == player && (d->flags & C_OUTPUTOFF))
      return 1;
  return 0;
}
